# -*- coding: utf-8 -*-
"""
---------AWS utilities--------------------------------------------------------
Usage:
AWS is class that defines all the configuration like access key or secret or pem key,
which can be used  both by ec2, s3, rds connections. At this point, it has been used for
ec2 and s3 connections.
AWS object can be populated from a file containing a json object
Eg: Use pem file to make ssh connection
    import socket, paramiko
    from util_aws import AWS
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.connect(hostname, 22)
    transport = paramiko.Transport(self.sock)
    transport.start_client()
    pemfile = AWS().v['AWS_KEYPEM'] # Get the private key pem file
    privatekey = paramiko.RSAKey.from_private_key_file(pemfile)
    transport.auth_publickey(username, privatekey)

Eg: Get the keypair stored in the AWS for running a command using SSH
    identity = AWS().ec2_keypair_get()  # Get the key pair
    cmdstr = "df -k | grep swap"
    swapspace = ssh_cmdrun(ipadress, identity, cmdstr)

Eg: Get the access and secret key for accessing AWS
    access, secret = AWS().aws_accesskey_get()
    conn = boto.connect_s3(access, secret)
    bucket = conn.get_bucket(bucket_name)

Eg: Create a conn for us-east-2
    east2-conn = AWS().aws_conn_create(region='us-east-2')

Eg: Create windows conn for us-east-1
    east1-win-conn = AWS().aws_conn_create_windows(aws_region='us-east-1')

Utility methods:
ssh_cmdrun(): Runs a specific command using ssh library called paramiko and pem key file.
  ipaddress = '12.30.24.21'
  identity = AWS().ec2_keypair_get()
  cmdstr = "ls -lrt"
  reverselisting = ssh_cmdrun(ipadress, identity, cmdstr)

For a instance type, fetch ram, cpu usage as an array for spot requests only
  ec2_instance_getallstate_cli('t2.small')

Each instance usage details
  spot_requestid = '<spot request id>'
  ipaddr = '12.30.15.25'
  cpuusage, usageram, totalram = ec2_instance_usage(spot_requestid, ipaddr)

Build a template for a spot request and store in default location(/tmp/ec_spot_config)
  ec2_config_build_template_cli('t2.medium')


Request spot instance.
  instance = 't3.small'
  spot_price = 0.59
  region = 'us-east-1'
  ec2_spot_start_cli(instance, spot_price, region)

Current list of spot instances.
  spot_list = ec2_spot_instance_list()


Stop the spot instance based on the instanceid
  spot_request_id = '<Req Id>'
  ec2_instance_stop(spot_request_id)

For a given instance type, run a shell script to make a http call
and grep the price for the same.
  instance_type = 't3.medium'
  ec2_get_spot_price(instance_type)




TODO :
  All in one file util_aws.py  : Top priority
 - Remove Dependance from util.py
 
 - Create/check  /USER/.aws/ folder to put all the HARD CODE values. 
    .ssh : for SSH keys.

 - AWS: the default values should be in a default file, not in code
        This default json file should be in the same directory as this file.

 - boto.config.get in lines 148 and 149 are not working, they need to be fixed.
 - Have imports at the top unless dependencies issues are to be resolved.
 
 Short name is better - AWS.v is not good descriptive name for a instance variable, could be
   AWS.constant_values or AWS.global_values.

 - Lines 206-212, optimize as EC2Connection(access, key, region=r), if it fails then check
    if it is a valid region. Anyway as of now 16 valid regions are there and they can be
    statically checked, why fetch them?

 - Dont see a difference between a  windows and a non-windows EC2 connection.

 - ssh_cmdrun, ssh_put  methods to be moved to a separate file, can be used by other files,
    which do not require other AWS functionality. Platform check for Windows machine and
    return a False. Also adapt it to run a list of commands as it saves making connection
    multiple times. In lines 366-388, 2 commands are run on the same host and the connections
    are made twice.
 - os.system does not return output, os.popen does return.  For all aws ec2 commands, it is
    json output, so always good to capture it.
 - aws_ec2_get_instanceid, this will need pagination support, an account may have more than
    one page of instances.
 - aws_ec2_allocate_elastic_ip and aws_ec2_allocate_eip are same code set, need a merge. Multiple
    times defined (Lines 535 and 746)



NG - json.load(<filename>) is notorius to fail, so have a json_utils.py and have a
        try-catch block with fallback of returning a empty dict.  

NG - Have all these constants in a separate file constants.py and import
   only required constants like
   from constants import AWS_ACCESS_LOCAL
   This way we bring structure to the code and anybody can update and maintain it.

NG -  Keep CLI and boto version separate.separate
      aws_ec2_spot_start and aws_ec2_spot_start_cli are same, one uses boto and the other aws ec2.
     the code should switch between these two calls based on configuration ( global or individual calls)
 -
"""

from __future__ import division, print_function

import csv
import json
import socket
import os
import re
import sys
import subprocess
import errno
import fnmatch
import shutil
import zipfile
from stat import S_ISDIR
from time import sleep


import boto
from boto import ec2
from boto.ec2.connection import EC2Connection
from boto.ec2.blockdevicemapping import BlockDeviceMapping, EBSBlockDeviceType
from boto.s3.connection import S3Connection

from tqdm import tqdm
import paramiko
import ntpath
import urllib.request


###############################################################################
def get_host_public_ipaddress():
    f = urllib.request.urlopen("http://ipinfo.io/ip")
    val = f.read()
    ip = str(val, 'utf-8').replace('\n', '') if val else ''
    f.close()
    return ip


def exists_file(fname):
    """Check if path exists and is a file"""
    if fname and os.path.exists(fname) and os.path.isfile(fname):
        return True
    return False


def exists_dir(dirname):
    """Check if this path exists and is a directory"""
    if dirname and os.path.exists(dirname) and os.path.isdir(dirname):
        return True
    return False


def json_from_file(jsonfile, defval=None):
    """ Get json data from the file."""
    jsondata = defval
    try:
        if exists_file(jsonfile):
            with open(jsonfile) as infile:
                data = infile.read()
                jsondata = json.loads(data)
    except Exception as ex:
        print('Failed to load json from file: %s, exception: %s' % (jsonfile, ex))
    return jsondata


###############################################################################
class AWS:
    """
    All the globals for AWS utility functionalities.
    """
    def __init__(self, name=None, keypair=None, keypem=None):
        """Nothing to be constructed """
        self.v = {
            "AWS_CONFIG_FOLDER": "/.aws/",
            "AWS_ACCESS_LOCAL": 'D:/_devs/keypair/aws_access.py',
            "AWS_KEYPEM": keypem if keypem else "D:/_devs/keypair/oregon/aws_ec2_oregon.pem",
            "AWS_KEYPAIR": keypair if keypair else 'aws_ec2_oregon',
            "AWS_REGION": 'us-west-2',
            "DEFAULT_INSTANCE_TYPE": 't3.small',
            "EC2CWD": '/home/ubuntu/',
            "EC2_CONN": None,
            "EC2_FILTERS": ('id', 'ip_address'),
            "EC2_ATTRIBUTES": (
                "id", "instance_type", "state", "public_dns_name", "private_dns_name",
                "state_code", "previous_state", "previous_state_code", "key_name",
                "launch_time", "image_id", "placement", "placement_group",
                "placement_tenancy", "kernel", "ramdisk", "architecture", "hypervisor",
                "virtualization_type", "product_codes", "ami_launch_index", "monitored",
                "monitoring_state", "spot_instance_request_id", "subnet_id", "vpc_id",
                "private_ip_address", "ip_address", "platform", "root_device_name",
                "root_device_type", "state_reason", "interfaces", "ebs_optimized",
                "instance_profile"
            ),
            "SPOT_CFG_FILE": ".aws/ec_spot_config"
        }
        
        if ".json" in name:
            dd = json_from_file(name, {})
        else:
            dd = json_from_file(os["HOME"] + "/.aws/" + name, {})
        if dd:
            self.v.update(dd)
        self.v = dict2(self.v)
        self.conn = None

    def aws_accesskey_get(self):
        """Return a tuple of AWS credentials (access key id and secret access key)"""
        # Try from Boto Config
        access = boto.config.get('Credentials', 'aws_access_key_id')
        key = boto.config.get('Credentials', 'aws_secret_access_key')

        if access and key:
            print(access, key)
            return access, key

        # Finally try the manual Config
        # with open(self.v['AWS_ACCESS_LOCAL']) as aws_file:
        #     dd = json.loads(aws_file.read())
        dd = json_from_file(self.v['AWS_ACCESS_LOCAL'], {})
        if 'AWS_ACCESS_KEY_ID' in dd:
            access = dd['AWS_ACCESS_KEY_ID']
        if 'AWS_SECRET_ACCESS_KEY' in dd:
            key = dd['AWS_SECRET_ACCESS_KEY']
        return access, key

    def ec2_keypair_get(self, keypair=""):
        sshdir = "%s/.ssh" % os.environ['HOME'] if 'HOME' in os.environ else '/home/ubuntu/'
        keypair = keypair if keypair else self.v['AWS_KEYPEM']
        identity = '%s/%s' % (sshdir, keypair)
        return identity

    def get_keypair(self):
        """Get the current keypair used"""
        return self.v['AWS_KEYPEM'], self.v['AWS_KEYPAIR']

    def set_keypair(self, keypairname, keypairlocation):
        """Set the keypair to be used."""
        if keypairname and keypairlocation:
            self.v['AWS_KEYPAIR'] = keypairname
            self.v['AWS_KEYPEM'] = keypairlocation

    @classmethod
    def set_attribute(cls, key, value):
        """Add or update attribute to the class, maybe protect with a lock."""
        setattr(cls, key, value)

    def get_ec2_conn(self):
        """Return the current EC2 connection."""
        return self.conn

    def aws_conn_create(self, region='', access='', key=''):
        """ EC2 connection to specified region or  AWS.v['AWS_REGION']"""
        if not region:
            region = self.v['AWS_REGION']
        if not access or not key:
            access, key = self.aws_accesskey_get()
        if not access or not key:
            return None
        # We could have used connection from any region.
        conn = EC2Connection(access, key, region=region)
        if conn:
            self.conn = conn
            return self.conn
        else:
            regions = aws_conn_getallregions(conn)
            for r in regions:
                if r.name == region:
                    self.conn = EC2Connection(access, key, region=r)
                    return self.conn
        print('Region not Found')
        return None

    def aws_conn_create_windows(self, aws_region=''):
        """ Return ec2_conn for windows system, otherwise none."""
        if not aws_region:
            aws_region = self.v['AWS_REGION']
        ec2_conn = self.conn
        if sys.platform.find('win') > -1 and not ec2_conn:
            # dd = {}
            # with open(self.v['AWS_ACCESS_LOCAL']) as aws_file:
            #     dd = json.loads(aws_file.read())
            dd = json_from_file(self.v['AWS_ACCESS_LOCAL'], {})
            if 'AWS_ACCESS_KEY_ID' in dd:
                access = dd['AWS_ACCESS_KEY_ID']
            if 'AWS_SECRET_ACCESS_KEY' in dd:
                key = dd['AWS_SECRET_ACCESS_KEY']
            if access and key:
                ec2_conn = boto.ec2.connect_to_region(aws_region, aws_access_key_id=access,
                                                      aws_secret_access_key=key)
                print(ec2_conn)
                # We should store this in AWS
                self.conn = ec2_conn
        return self.conn


###############StandAlone access ###################################################################
def ssh_cmdrun(hostname, key_file, cmdstr, remove_newline=True, isblocking=True):
  """ Make an ssh connection using paramiko and  run the command
   http://sebastiandahlgren.se/2012/10/11/using-paramiko-to-send-ssh-commands/
   https://gist.github.com/kdheepak/c18f030494fea16ffd92d95c93a6d40d
   https://github.com/paramiko/paramiko/issues/501
   https://unix.stackexchange.com/questions/30400/execute-remote-commands-completely-detaching-from-the-ssh-connection
   
  """
  try:
    # print('%s: %s' % ('*' * 10, key_file))
    ssh = paramiko.SSHClient()
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    # k = paramiko.RSAKey.from_private_key_file(key_file)
    # ssh.connect(hostname, pkey=k, timeout=5, look_for_keys=False)
    ssh.connect(hostname, key_filename=key_file, timeout=5, look_for_keys=False)

    stdin, stdout, stderr = ssh.exec_command(cmdstr) #No Blocking  , get_pty=False
    
    """
    if not isblocking :
       # Buggy code, use Screen instead
       sleep(10) # To let run the script
       ssh.close()
       return None
    """
    
    #### Can be Blocking for long running process  screen -d -m YOURBASH
    data  = stdout.readlines()  #Blocking code
    value = ''.join(data).replace('\n', '') if remove_newline else ''.join(data)
    
    err_msg = stderr.readlines() 
    if len(err_msg) > 0 :
      print(err_msg  )
    
    ssh.close()
    return value
  except Exception as e :
    print("Error Paramiko: %s" % hostname, e)
    return None


def ssh_put(hostname, key_file, remote_file, msg=None, filename=None):
    """ Make an ssh connection using paramiko and  run the command
     http://sebastiandahlgren.se/2012/10/11/using-paramiko-to-send-ssh-commands/
     https://gist.github.com/kdheepak/c18f030494fea16ffd92d95c93a6d40d
 
     https://github.com/paramiko/paramiko/issues/501
    """
    ssh = paramiko.SSHClient()
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    ssh.connect(hostname, key_filename=key_file, timeout=5)
    #stdin, stdout, stderr = ssh.exec_command(cmdstr, get_pty=False) #No Blocking 
    
    if filename is not None :
      msg = open(filename,mode="r").readlines()
    
    ftp = ssh.open_sftp()
    file=ftp.file(remote_file, "a", -1)
    file.write(msg)
    file.flush()
    ftp.close()
    ssh.close()


def ec2_instance_getallstate_cli(default_instance_type="t3.medium"):
  """
  use to update the global INSTANCE_DICT
  "id" :  instance_type, ip_address, cpu, ram, cpu_usage, ram_usage
  """
  val = {}
  spot_list = ec2_spot_instance_list()
  spot_instances = []
  for spot_instance in spot_list['SpotInstanceRequests']:
    if re.match(spot_instance['State'], 'active', re.I) and \
      'InstanceId' in spot_instance:
        spot_instances.append(spot_instance['InstanceId'])
  # print(spot_instances)
  
  for spot in spot_instances:
    cmdargs = ['aws', 'ec2', 'describe-instances', '--instance-id', spot]
    cmd     = ' '.join(cmdargs)
    value   = os.popen(cmd).read()
    inst    = json.loads(value)
    ncpu    = 0
    ipaddr  = None
    instance_type = default_instance_type
    if inst and 'Reservations' in inst and inst['Reservations']:
      reserves = inst['Reservations'][0]
      if 'Instances' in reserves and reserves['Instances']:
        instance = reserves['Instances'][0]
        if 'CpuOptions' in instance and 'CoreCount' in instance['CpuOptions']:
          ncpu = instance['CpuOptions']['CoreCount']
        if 'PublicIpAddress' in instance and instance['PublicIpAddress']:
          ipaddr = instance['PublicIpAddress']
        instance_type = instance['InstanceType']
    
    if ipaddr:
      cpuusage, usageram, totalram = ec2_instance_usage(spot, ipaddr)
      # print(cpuusage, usageram, totalram)
      val[spot] = {
        'id': spot,
        'instance_type': instance_type,
        'cpu': ncpu,
        'ip_address': ipaddr,
        'ram': totalram,
        'cpu_usage': cpuusage,
        'ram_usage': usageram
      }
  # print(val)
  return val


################################################################################ 
def ec2_instance_usage(instance_id=None, ipadress=None):
  """
  https://stackoverflow.com/questions/20693089/get-cpu-usage-via-ssh
  https://haloseeker.com/5-commands-to-check-memory-usage-on-linux-via-ssh/
  """
  cpuusage, ramusage, totalram = [None] * 3
  totalram = None
  if instance_id and ipadress:
    identity = AWS().ec2_keypair_get()
    cmdstr = "top -b -n 10 -d.2 | grep 'Cpu' | awk 'BEGIN{val=0.0}{ if( $2 > val ) val = $2} END{print(val)}'"
    cpuusage = ssh_cmdrun(ipadress, identity, cmdstr)
    cpuusage = 100.0  if not cpuusage else float(cpuusage)
    cmdstr = "free | grep Mem | awk '{print $3/$2 * 100.0, $2}'"
    ramusage = ssh_cmdrun(ipadress, identity, cmdstr)
    if not ramusage:
        totalram = 0
        usageram = 100.0
    else:
        vals = ramusage.split()
        usageram = float(vals[0]) if vals and vals[0] else 100.0
        totalram = int(vals[1]) if vals and vals[1] else 0
  return cpuusage, usageram, totalram


def ec2_config_build_template_cli(instance_type, amiId=None, keypair=None, 
         default_instance_type=None, spot_cfg_file=None):
  """ Build the spot json config into a json file. """
  spot_config = {
    "ImageId": amiId,
    "KeyName": keypair, 
    "SecurityGroupIds": ["sg-4b1d6631", "sg-42e59e38"],        
    "InstanceType": instance_type if instance_type else default_instance_type,
    "IamInstanceProfile": {
      "Arn": "arn:aws:iam::013584577149:instance-profile/ecsInstanceRole"
    },
    "BlockDeviceMappings": [
      {
        "DeviceName": "/dev/sda1",
        "Ebs": {
          "DeleteOnTermination": True,
          "VolumeSize": 60
        }                      
      }
    ]
  }
  if not spot_cfg_file:
      spot_cfg_file = AWS().v['SPOT_CFG_FILE']
  with open(spot_cfg_file, 'w') as spot_file:
    spot_file.write(json.dumps(spot_config))


def ec2_spot_start_cli(instance_type, spot_price, region=None, waitsec=100):
  """
  Request a spot instance based on the price for the instance type
  # Need a check if this request has been successful.
  
  100 sec to be provisionned and started.
  """
  aws = AWS()
  if not instance_type:
      instance_type = aws.v['DEFAULT_INSTANCE_TYPE']
  if not region:
      region = aws.v['AWS_REGION']
  ec2_config_build_template_cli(instance_type)
  cmdargs = [
    'aws', 'ec2', 'request-spot-instances',
    '--region', region,
    '--spot-price', str(spot_price),
    '--instance-count', "1",
    ' --type', 'one-time',
    '--launch-specification', 'file://%s' % aws.v['SPOT_CFG_FILE']
  ]
  print(cmdargs)
  cmd = ' '.join(cmdargs)
  os.system(cmd)
  sleep(waitsec)  # It may not be fulfilled in 50 secs.
  ll= ec2_spot_instance_list()
  return ll['SpotInstanceRequests'] if 'SpotInstanceRequests' in ll else []


def ec2_spot_instance_list():
  """ Get the list of current spot instances. """
  cmdargs = [
    'aws', 'ec2', 'describe-spot-instance-requests'
  ]
  cmd = ' '.join(cmdargs)
  value = os.popen(cmd).read()
  instance_list = json.loads(value)
  return instance_list


def ec2_instance_stop(instance_list) :
  """ Stop the spot instances ainstances u stop any other instance, this should work"""
  instances = instance_list
  if instances:
    if isinstance(instance_list, list) :
        instances = ','.join(instance_list)
    cmdargs = [
      'aws', 'ec2', 'terminate-instances', '--instance-ids', instances
    ]
    cmd = ' '.join(cmdargs)
    os.system(cmd)
    return instances.split(",")


def ec2_get_spot_price(instance_type):
  """ Get the spot price for instance type in us-west-2"""
  value = 0.0
  if os.path.exists('./aws_spot_price.sh') and os.path.isfile('./aws_spot_price.sh'):
    cmdstr = "./aws_spot_price.sh %s | grep Price | awk '{print $2}'" % instance_type
    value = os.popen(cmdstr).read()
    value = value.replace('\n', '') if value else 0.10
  return tofloat(value)


def aws_conn_getallregions(conn=None):
    """ Get the regions for the connection passed. """
    if conn:
        return conn.get_all_regions()
    return []  # Fail safe, caller may not check.


def aws_conn_getinfo(conn):
    """ Connection region name """
    region_name = None
    if conn:
        print(conn.region.name)
        region_name = conn.region.name
    return region_name


def aws_ec2_ami_create(conn, ip_address='', ami_name=''):
    """ Createan an AMI for the instance represented by the ipadress"""
    if conn and ip_address:
        instance_list = conn.get_all_instances(
            filters={
                "ip_address": ip_address
            }
        )
        instance = instance_list[0].instances[0] if instance_list else None
        if instance:
            ami_id = instance.create_image(ami_name)
            print("Image creation(%s) for AMI ID %s started...." % (ami_name, ami_id))
            return True
    return False


def aws_ec2_get_instanceid(conn, filters=None):
    """ Get Instance Id based on the filters. """
    if not filters:
        filters = {
            "ip_address": ""
        }
    if conn:
        instance_list = conn.get_all_instances(filters=filters)
        instance = instance_list[0].instances[0] if instance_list else None
        if instance:
            return instance.id
    return None


def aws_ec2_allocate_elastic_ip(conn, instance_id='', elastic_ip='', region=''):
    """ Allocate elastic IP in the region."""
    if not region:
        region = AWS().v['APNORTHEAST2']
    if conn and instance_id:
        if not elastic_ip:
            eip = conn.allocate_address()
            elastic_ip = eip.public_ip
        if elastic_ip:
            conn.associate_address(instance_id=instance_id, public_ip=elastic_ip)
            print('Elastic assigned Public IP: %s,Instance_ID: %s' % (elastic_ip, instance_id))
            return elastic_ip
    return None


def aws_ec2_allocate_eip(instance_id, conn=None, eip_allocation_id=None, eip_public_ip=None,
                         allow_reassociation=False):
    """
    Assign an Elastic IP to an instance. Works with either by specifying the
    connection: EC2Connection object
    instance_id: Desired EC2 instance's ID
    eip_allocation_id: ID of Elastic IP to assign (Required if no public_ip)
    eip_public_ip: Elastic IP to assign (Required if no allocation_id)
    allow_reassociation: Option to turn off reassociation (check caveats below)
    """
    if conn:
        if eip_public_ip:
            conn.associate_address(instance_id=instance_id, public_ip=eip_public_ip,
                                   allow_reassociation=allow_reassociation)
            return True
        elif eip_allocation_id:
            conn.associate_address(instance_id=instance_id, allocation_id=eip_allocation_id,
                                   allow_reassociation=allow_reassociation)
            return True
        else:
            raise ValueError("eip_public_ip and eip_allocation_id cannot be both None!")


def aws_ec2_spot_start(conn, region, key_name="ecsInstanceRole", inst_type="cx2.2", ami_id="",
                       pricemax=0.15, elastic_ip='', pars=None):
    """
    :param conn: Connector to Boto
    :param region: AWS region (us-east-1,..)
    :param key_name: AWS SSH Key Name(in EC2 webspage )
    :param security_group: AWS security group id
    :param inst_type:  AWS EC2 instance type(t1.micro, m1.small ...)
    :param ami_id: AWS AMI ID
    :param pars: Disk Size, Volume type (General Purpose SSD - gp2, Magnetic etc)
    :param pricemax: minmum spot instance bid price
    """
    if not conn:
        return None
    if not pars:
        pars = {
            "security_group": [""],
            "disk_size": 25,
            "disk_type": "ssd",
            "volume_type": "gp2"
        }
    pars = dict2(pars)   #Dict to Attribut Dict
    print("starting EC2 Spot Instance")

    try:
        block_map = BlockDeviceMapping()
        device = EBSBlockDeviceType()
        device.size = int(pars.disk_size)  # size in GB
        device.volume_type = pars.volume_type  # [ standard | gp2 | io1 ]
        device.delete_on_termination = False
        block_map["/dev/xvda"] = device
        print("created a block device")

        req = conn.request_spot_instances(price=pricemax, image_id=ami_id, instance_type=inst_type,
                                          key_name=key_name, security_groups=pars.security_group,
                                          block_device_map=block_map)
        print("Spot instance request created, status: %s" % (req[0].status))
        print("Waiting for spot instance provisioning")
        while True:
            current_req = conn.get_all_spot_instance_requests([req[0].id])[0]
            if current_req.state == "active":
                print("Spot instance provisioning successful.")
                instance = conn.get_all_instances([current_req.instance_id])[0].instances[0]
                aws_ec2_allocate_elastic_ip(conn, current_req.instance_id, elastic_ip, region)
                # Print Instance details : ID, IP adress
                print('IP Addess:%s, instance_id: %s' % (getattr(instance, "ip_address"),
                                                         getattr(instance, 'id')))
                return instance
            # print('.', end='')
            sleep(15)
    except:
        pass


def aws_ec2_spot_stop(conn, ipadress="", instance_id=""):
    """
    :param conn: connector
    :param ipadress:   of the instance  to Identify the instance.
    :param instance_id:  OR use instance ID....
    :return:
    """
    if not conn:
        return False
    if not instance_id:
        instance_id=  aws_ec2_get_instanceid(conn, ipadress)  # Get ID from IP Adresss
    try:
        print("Terminating Spot Instance : %s" % (str(instance_id)))
        conn.terminate_instances(instance_ids = [instance_id])
        print ("Terminating spot instance Successful!")
        return True
    except Exception as e:
        print ("Error : Failed to terminate. %s" % (str(e)))
    return False


def aws_ec2_res_start(conn, region, key_name, ami_id, inst_type="cx2.2",
                      min_count =1, max_count=1, pars=None):
    """
    normal instance start
    :param conn: Connector to Boto
    :param region: AWS region (us-east-1,..)
    :param key_name: AWS  SSH Key Name
    :param security_group: AWS security group id
    :param inst_type:  AWS EC2 instance type (t1.micro, m1.small ...)
    :param ami_id:  AWS AMI ID
    :param min_count: Minumum number of instances
    :param max_count : Maximum number of instances
    :param pars: Disk Size, Volume type (General Purpose SSD - gp2,Magnetic etc)
    :return
    """
    if not conn:
        return None
    if not pars:
        pars = {
            "security_group": [""],
            "disk_size": 25,
            "disk_type": "ssd",
            "volume_type": "gp2"
        }
    pars = dict2(pars)   #Dict to Attribut Dict
    try: 
        block_map = BlockDeviceMapping()
        device = EBSBlockDeviceType()
        device.size = int(pars.disk_size)  # size in GB
        device.volume_type = pars.volume_type  # [ standard | gp2 | io1 ]
        device.delete_on_termination = False
        block_map["/dev/xvda"] = device
        print("created a block device")
        req = conn.run_instances(image_id=ami_id, min_count=min_count,
                                 max_count=max_count,
                                 instance_type=inst_type, key_name=key_name,
                                 security_groups= pars.security_group,
                                 block_device_map=block_map)
        instance_id = req.instances[0].id
        print("EC2 instance has been created. Instance ID : %s" % instance_id)
        print("Waiting for EC2 instance provisioning")
        while True:
            print('.', end='')
            current_req = conn.get_all_instances(instance_id)
            if current_req[0].instances[0].state.lower() == "running":
                print("EC2 instance provisioning successful and the instance is running.")
                aws_ec2_printinfo(current_req[0].instances[0])
                return current_req[0].instances[0]
            print('.'),
            sleep(30)
    except Exception as e:
        print("Error : %s " % (str(e)))
    return None


def aws_ec2_res_stop(conn, ipadress="", instance_id=""):
    """
    :param conn: connector
    :param ipadress:     Of the instance  to Identify the instance.
    :param instance_id:  OR use instance ID....
    """
    if not conn:
        return False
    if not instance_id:
        instance_id = aws_ec2_get_instanceid(conn, ipadress)
    try:
        print("Stopping EC2 Instance : %s" % (str(instance_id)))
        req = conn.stop_instances(instance_ids=[instance_id])
        print("EC2 instance has been stopped successfully. %s" % req)
        return True
    except Exception as e:
        print("Error : Failed to stop EC2 instance. %s" % str(e))
    return False


def aws_ec2_printinfo(instance=None, ipadress="", instance_id=""):
    """
    Idenfiy instnance of
    :param instance: ipadress
    :param instance_id:
    :return: return info on the instance : ip, ip_adress,
    """
    if not ipadress:
        print("")
    if not instance_id:
        pass


def aws_ec2_get_folder(ipadress, fromfolder1, tofolder1):
    """Copy Folder from remote to  EC2 folder"""
    pass


def aws_s3_url_split(url):
    """Split into Bucket, url."""
    url_parts = url.split("/")
    return url_parts[0], "/".join(url_parts[1:])


def aws_s3_getbucketconn(s3dir):
    """ Get S3 bucket. """
    # import boto.s3
    bucket_name, todir = aws_s3_url_split(s3dir)
    access, secret = AWS().aws_accesskey_get()
    conn = boto.connect_s3(access, secret)
    bucket = conn.get_bucket(bucket_name)  # location=boto.s3.connection.Location.DEFAULT)
    return bucket


def aws_s3_put(fromdir_file='dir/file.zip', todir='bucket/folder1/folder2'):
    """
    Copy File or Folder to S3
    "aws s3 cp s3://s3-bucket/scripts/HelloWorld.sh /home/ec2-user/HelloWorld.sh",
    "chmod 700 /home/ec2-user/HelloWorld.sh",
    "/home/ec2-user/HelloWorld.sh"
    """
    def percent_cb(complete, total):
        sys.stdout.write('.'); sys.stdout.flush()

    # import boto.s3
    bucket = aws_s3_getbucketconn(todir)
    bucket_name, todir = aws_s3_url_split(todir)

    MAX_SIZE = 20 * 1000 * 1000
    PART_SIZE = 6 * 1000 * 1000

    if fromdir_file.find('.') > -1:   # Case of Single File
        filename = os_file_getname(fromdir_file)
        fromdir_file = os_file_getpath(fromdir_file) + '/'
        uploadFileNames = [filename]
    else:
        uploadFileNames = []
    for (fromdir_file, dirname, filename) in os.walk(fromdir_file):
        uploadFileNames.extend(filename)
        break

    for filename in uploadFileNames:
        sourcepath = os.path.join(fromdir_file + filename)
        destpath = os.path.join(todir, filename)
        print("Uploading %s to Amazon S3 bucket %s" % (sourcepath, bucket_name))
        filesize = os.path.getsize(sourcepath)
        if filesize > MAX_SIZE:
            print("multipart upload")
            mp = bucket.initiate_multipart_upload(destpath)
            fp = open(sourcepath, 'rb')
            fp_num = 0
            while fp.tell() < filesize:
                fp_num += 1
                print("uploading part %i" %fp_num)
                mp.upload_part_from_file(fp, fp_num, cb=percent_cb, num_cb=10,
                                         size=PART_SIZE)
            mp.complete_upload()
        else:
            print("singlepart upload: " + fromdir_file + ' TO ' + todir)
            k = boto.s3.key.Key(bucket)
            k.key = destpath
            k.set_contents_from_filename(sourcepath, cb=percent_cb, num_cb=10)


def aws_s3_get(froms3dir='task01/', todir='', bucket_name='zdisk'):
    """Get from S3 file/folder"""
    bucket_name, dirs3 = aws_s3_url_split(froms3dir)
    bucket = aws_s3_getbucketconn(froms3dir)
    bucket_list = bucket.list(prefix=dirs3)  # /DIRCWD/dir2/dir3

    for l in bucket_list:
        key1 = str(l.key)
        file1 = os_file_getname(key1)
        path2 = os_file_getpath(key1)
        # Remove prefix path of S3 to machine
        path1 = os.path.relpath(path2, dirs3).replace('.', '')
        d = todir + '/' + path1
        if not os.path.exists(d):
            os.makedirs(d)
        try:
            l.get_contents_to_filename(d+ '/'+ file1)
        except OSError:
            pass


def aws_s3_folder_printtall(bucket_name='zdisk'):
    """ Print all S3 bucket folders"""
    access, secret = AWS().aws_accesskey_get()
    conn = boto.connect_s3(access, secret)
    bucket = conn.create_bucket(bucket_name,
                                location=boto.s3.connection.Location.DEFAULT)
    folders = bucket.list("", "/")
    for folder in folders:
        print(folder.name)
    return folders


def aws_s3_file_read(bucket1, filepath):
    """
    s3_client = boto3.client('s3')
    # Download private key file from secure S3 bucket
    s3_client.download_file('s3-key-bucket','keys/keyname.pem',
                            '/tmp/keyname.pem')
    """
    # from boto.s3.connection import S3Connection
    conn = S3Connection(AWS().aws_accesskey_get())
    response = conn.get_object(Bucket=bucket1, Key=filepath)
    file1 = response["Body"]
    return file1


class aws_ec2_ssh(object):
    """
    ssh= aws_ec2_ssh(host)
    print ssh.command('ls ')
    ssh.put_all( DIRCWD +'linux/batch/task/elvis_prod_20161220', EC2CWD + '/linux/batch/task' )
    ssh.get_all(  EC2CWD + '/linux/batch/task',  DIRCWD +'/zdisks3/fromec2' )
    # Detects DSA or RSA from key_file, either as a string filename or a file object.
    # Password auth is possible, but I will judge you for
    # ssh=SSHSession('targetserver.com','root',key_file=open('mykey.pem','r'))
    # ssh=SSHSession('targetserver.com','root',key_file='/home/me/mykey.pem')
    # ssh=SSHSession('targetserver.com','root','mypassword')
    # ssh.put('filename','/remote/file/destination/path')
    # ssh.put_all('/path/to/local/source/dir','/path/to/remote/destination')
    # ssh.get_all('/path/to/remote/source/dir','/path/to/local/destination')
    # ssh.command('echo "Command to execute"')
    """

    def __init__(self, hostname, username='ubuntu', key_file=None, password=None):
        # import socket
        # Accepts a file-like object (anything with a readlines() function)
        # in either dss_key or rsa_key with a private key.  Since I don't
        # ever intend to leave a server open to a password auth.
        self.host = hostname
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.connect((hostname,22))
        self.t = paramiko.Transport(self.sock)
        self.t.start_client()
        # keys = paramiko.load_host_keys(os.path.expanduser('~/.ssh/known_hosts'))
        # key = self.t.get_remote_server_key()
        # supposed to check for key in keys, but I don't much care right now to
        # find the right notation

        key_file = AWS().v['AWS_KEYPEM'] if not key_file else key_file
        pkey = paramiko.RSAKey.from_private_key_file(key_file)
        self.t.auth_publickey(username, pkey)
        self.sftp = paramiko.SFTPClient.from_transport(self.t)
        print(self.command('ls '))

    def command(self, cmd):
        # Breaks the command by lines, sends and receives
        # each line and its output separately
        # Returns the server response text as a string
        chan = self.t.open_session()
        chan.get_pty()
        chan.invoke_shell()
        chan.settimeout(20.0)
        ret = ''
        try:
            ret += chan.recv(1024)
        except:
            chan.send('\n')
            ret += chan.recv(1024)
        for line in cmd.split('\n'):
            chan.send(line.strip() + '\n')
            ret += chan.recv(1024)
        return ret

    def cmd(self, cmdss):
        ss = self.command(cmdss)
        print(ss)

    def put(self, localfile, remotefile):
        #  Copy localfile to remotefile, overwriting or creating as needed.
        self.sftp.put(localfile, remotefile)

    def put_all(self, localpath, remotepath):
        # recursively upload a full directory
        # localpath= localpath[:-1] if localpath[-1]=='/' else localpath
        # remotepath= remotepath[:-1] if remotepath[-1]=='/' else remotepath
        os.chdir(os.path.split(localpath)[0])
        parent = os.path.split(localpath)[1]
        print(parent)
        for walker in os.walk(parent):
            try:
                self.sftp.mkdir(os.path.join(remotepath, walker[0]).replace('\\', "/"))
            except:
                pass
            for file in walker[2]:
                print(os.path.join(walker[0], file).replace('\\', '/').replace('\\', '/'),
                      os.path.join(remotepath, walker[0], file).replace('\\', '/'))
                self.put(os.path.join(walker[0], file).replace('\\', '/'),
                         os.path.join(remotepath, walker[0], file).replace('\\', '/'))

    def get(self, remotefile, localfile):
        #  Copy remotefile to localfile, overwriting or creating as needed.
        self.sftp.get(remotefile, localfile)

    def sftp_walk(self, remotepath):
        # Kindof a stripped down  version of os.walk, implemented for
        # sftp.  Tried running it flat without the yields, but it really
        # chokes on big directories.
        path = remotepath
        files = []
        folders = []
        for f in self.sftp.listdir_attr(remotepath):
            if S_ISDIR(f.st_mode):
                folders.append(f.filename)
            else:
                files.append(f.filename)
        print(path, folders, files)
        yield path, folders, files
        for folder in folders:
            new_path = os.path.join(remotepath, folder).replace('\\', '/')
            for x in self.sftp_walk(new_path):
                yield x

    def get_all(self, remotepath, localpath):
        #  recursively download a full directory
        #  Harder than it sounded at first, since paramiko won't walk
        # For the record, something like this would gennerally be faster:
        # ssh user@host 'tar -cz /source/folder' | tar -xz
        localpath = localpath[:-1] if localpath[-1] == '/' else localpath
        remotepath = remotepath[:-1] if remotepath[-1] == '/' else remotepath

        self.sftp.chdir(os.path.split(remotepath)[0])
        parent = os.path.split(remotepath)[1]
        try:
            os.mkdir(localpath)
        except:
            pass
        for walker in self.sftp_walk(parent):
            try:
                os.mkdir(os.path.join(localpath, walker[0]).replace('\\', '/'))
            except:
                pass
            for file in walker[2]:
                print(os.path.join(walker[0], file).replace('\\', '/'),
                      os.path.join(localpath, walker[0], file).replace('\\', '/'))
                self.get(os.path.join(walker[0], file).replace('\\', '/'),
                         os.path.join(localpath, walker[0], file).replace('\\', '/'))

    def write_command(self, text, remotefile):
        # Writes text to remotefile, and makes remotefile executable.
        # This is perhaps a bit niche, but I was thinking I needed it.
        # For the record, I was incorrect.
        self.sftp.open(remotefile, 'w').write(text)
        self.sftp.chmod(remotefile, 755)

    def python_script(self, ipython_path='/home/ubuntu/anaconda3/bin/ipython ',
                      script_path="", args1=""):
        cmd1 = ipython_path + ' ' + script_path + ' ' + '"' + args1 +'"'
        self.cmd2(cmd1)
        # self.command(cmd1)

    def command_list(self, cmdlist):
        for command in cmdlist:
            print("Executing {}".format(command))
            ret = self.command(command)
            print(ret)
        print('End of SSH Command')

    def listdir(self, remotedir):
        return self.sftp.listdir(remotedir)

    def jupyter_kill(self):
        pid_jupyter = aws_ec2_ssh_cmd(cmdlist=['fuser 8888/tcp'], host=self.host,
                                      doreturn=1)[0][0].strip()
        print(self.command('kill -9 ' + pid_jupyter))

    def jupyter_start(self):
        pass

    def cmd2(self, cmd1):
        return aws_ec2_ssh_cmd(cmdlist=[cmd1], host=self.host, doreturn=1)

    def _help_ssh(self):
        s = """
           fuser 8888/tcp     Check if Jupyter is running
           ps -ef | grep python     :List of  PID Python process
           kill -9 PID_number     (i.e. the pid returned)
           top: CPU usage
          """
        print(s)

    def put_all_zip(self, suffixfolder="", fromfolder="", tofolder="",
                    use_relativepath=True, usezip=True, filefilter="*.*",
                    directorylevel=1, verbose=0):
        """
        fromfolder:  c:/folder0/folder1/folder2/*
        suffixfolder: /folders1/folder2/
        tofolder: /home/ubuntu/myproject1/
        """
        fromfolder = fromfolder if fromfolder[-1] != '/' else  fromfolder[:-1]
        tmpfolder = '/'.join(fromfolder.split('/')[:-1]) + "/"
        zipname = fromfolder.split('/')[-1] + ".zip"
        print(tmpfolder, zipname)
        zipath = tmpfolder + zipname

        print("zipping")
        filezip = os_zipfolder(dir_tozip=fromfolder, zipname=zipath, dir_prefix=True,
                               iscompress=True)
        print("ssh on remote")
        remote_zipath = os.path.join(tofolder, zipname).replace('\\', '/')
        self.put(filezip, remote_zipath)

        if usezip:
            print("Unzip on remote " + self.host)
            cmd = '/usr/bin/unzip ' + remote_zipath + ' -d ' + tofolder
            ss = self.command(cmd)
            if verbose:
                print(ss)

def aws_ec2_ssh_create_con(contype='sftp/ssh', host='ip', port=22,
                           username='ubuntu', keyfilepath='', password='',
                           keyfiletype='RSA', isprint=1):
    """
    Transfert File  host = '52.79.79.1'
    keyfilepath = 'D:/_devs/aws/keypairs/ec2_instanc'
    # List files in the default directory on the remote computer.
    dirlist = sftp.listdir('.')
    sftp.get('remote_file.txt', 'downloaded_file.txt')
    sftp.put('testfile.txt', 'remote_testfile.txt')
    http://docs.paramiko.org/en/2.1/api/sftp.html
    """
    # import paramiko
    sftp, ssh, transport = None, None, None
    try:
        if not keyfilepath:
            keyfilepath = AWS().v['AWS_KEYPEM']
        if keyfiletype == 'DSA':
            key = paramiko.DSSKey.from_private_key_file(keyfilepath)
        else:
            key = paramiko.RSAKey.from_private_key_file(keyfilepath)

        if contype == 'sftp':
            # Create Transport object using supplied method of authentication.
            transport = paramiko.Transport((host, port))
            transport.add_server_key(key)
            transport.connect(None, username, pkey=key)
            sftp = paramiko.SFTPClient.from_transport(transport)
            if isprint:
                print('Root Directory : %s\n' % sftp.listdir())
            return sftp

        if contype == 'ssh':
            ssh = paramiko.SSHClient()
            ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            ssh.connect(hostname=host, username=username, pkey=key)

            # Test
            if isprint:
                stdin, stdout, stderr = ssh.exec_command("uptime;ls -l")
                stdin.flush()   #Execute
                data = stdout.read().splitlines()   #Get data
                print('Test Print Directory ls :')
                for line in data: print(line)
                return ssh

    except Exception as e:
        print('An error occurred creating client: %s: %s' % (e.__class__, e))
        if sftp: sftp.close()
        if transport: transport.close()
        if ssh: ssh.close()
    return None


def aws_ec2_ssh_cmd(cmdlist=["ls "], host='ip', doreturn=0, ssh=None,
                    username='ubuntu', keyfile=''):
    """
    SSH Linux terminal Command
    ### PEM File is needed
    aws_ec2_ssh_cmd(cmdlist=["ls " ], host='52.26.181.200', doreturn=1,
                    username='ubuntu', keyfilepath='')
    fuser 8888/tcp - Check if Jupyter is running
    ps -ef | grep python : List of  PID Python process
    kill -9 PID_number (i.e. the pid returned)
    top: CPU usage

    Run nohup python bgservice.py & to get the script to ignore the hangup signal and keep running.
    Output will be put in nohup.out.
    https://aws.amazon.com/blogs/compute/scheduling-ssh-jobs-using-aws-lambda/
    """


    def command(self, cmd):
        # Breaks the command by lines, sends and receives
        # each line and its output separately
        # Returns the server response text as a string
        chan = self.t.open_session()
        chan.get_pty()
        chan.invoke_shell()
        chan.settimeout(20.0)
        ret = ''
        try:
            ret += chan.recv(1024)
        except:
            chan.send('\n')
            ret += chan.recv(1024)
        for line in cmd.split('\n'):
            chan.send(line.strip() + '\n')
            ret += chan.recv(1024)
        return ret

    def cmd(self, cmdss):
        ss = self.command(cmdss)
        print(ss)

    def put(self, localfile, remotefile):
        #  Copy localfile to remotefile, overwriting or creating as needed.
        self.sftp.put(localfile, remotefile)

    def put_all(self, localpath, remotepath):
        # recursively upload a full directory
        # localpath= localpath[:-1] if localpath[-1]=='/' else localpath
        # remotepath= remotepath[:-1] if remotepath[-1]=='/' else remotepath
        os.chdir(os.path.split(localpath)[0])
        parent = os.path.split(localpath)[1]
        print(parent)
        for walker in os.walk(parent):
            try:
                self.sftp.mkdir(os.path.join(remotepath, walker[0]).replace('\\', "/"))
            except:
                pass
            for file in walker[2]:
                print(os.path.join(walker[0], file).replace('\\', '/').replace('\\', '/'),
                      os.path.join(remotepath, walker[0], file).replace('\\', '/'))
                self.put(os.path.join(walker[0], file).replace('\\', '/'),
                         os.path.join(remotepath, walker[0], file).replace('\\', '/'))

    def get(self, remotefile, localfile):
        #  Copy remotefile to localfile, overwriting or creating as needed.
        self.sftp.get(remotefile, localfile)

    def sftp_walk(self, remotepath):
        # Kindof a stripped down  version of os.walk, implemented for
        # sftp.  Tried running it flat without the yields, but it really
        # chokes on big directories.
        path = remotepath
        files = []
        folders = []
        for f in self.sftp.listdir_attr(remotepath):
            if S_ISDIR(f.st_mode):
                folders.append(f.filename)
            else:
                files.append(f.filename)
        print(path, folders, files)
        yield path, folders, files
        for folder in folders:
            new_path = os.path.join(remotepath, folder).replace('\\', '/')
            for x in self.sftp_walk(new_path):
                yield x

    def get_all(self, remotepath, localpath):
        #  recursively download a full directory
        #  Harder than it sounded at first, since paramiko won't walk
        # For the record, something like this would gennerally be faster:
        # ssh user@host 'tar -cz /source/folder' | tar -xz
        localpath = localpath[:-1] if localpath[-1] == '/' else localpath
        remotepath = remotepath[:-1] if remotepath[-1] == '/' else remotepath

        self.sftp.chdir(os.path.split(remotepath)[0])
        parent = os.path.split(remotepath)[1]
        try:
            os.mkdir(localpath)
        except:
            pass
        for walker in self.sftp_walk(parent):
            try:
                os.mkdir(os.path.join(localpath, walker[0]).replace('\\', '/'))
            except:
                pass
            for file in walker[2]:
                print(os.path.join(walker[0], file).replace('\\', '/'),
                      os.path.join(localpath, walker[0], file).replace('\\', '/'))
                self.get(os.path.join(walker[0], file).replace('\\', '/'),
                         os.path.join(localpath, walker[0], file).replace('\\', '/'))

    def write_command(self, text, remotefile):
        # Writes text to remotefile, and makes remotefile executable.
        # This is perhaps a bit niche, but I was thinking I needed it.
        # For the record, I was incorrect.
        self.sftp.open(remotefile, 'w').write(text)
        self.sftp.chmod(remotefile, 755)

    def python_script(self, ipython_path='/home/ubuntu/anaconda3/bin/ipython ',
                      script_path="", args1=""):
        cmd1 = ipython_path + ' ' + script_path + ' ' + '"' + args1 +'"'
        self.cmd2(cmd1)
        # self.command(cmd1)

    def command_list(self, cmdlist):
        for command in cmdlist:
            print("Executing {}".format(command))
            ret = self.command(command)
            print(ret)
        print('End of SSH Command')

    def listdir(self, remotedir):
        return self.sftp.listdir(remotedir)

    def jupyter_kill(self):
        pid_jupyter = aws_ec2_ssh_cmd(cmdlist=['fuser 8888/tcp'], host=self.host,
                                      doreturn=1)[0][0].strip()
        print(self.command('kill -9 ' + pid_jupyter))

    def jupyter_start(self):
        pass

    def cmd2(self, cmd1):
        return aws_ec2_ssh_cmd(cmdlist=[cmd1], host=self.host, doreturn=1)

    def _help_ssh(self):
        s = """
           fuser 8888/tcp     Check if Jupyter is running
           ps -ef | grep python     :List of  PID Python process
           kill -9 PID_number     (i.e. the pid returned)
           top: CPU usage
          """
        print(s)

    def put_all_zip(self, suffixfolder="", fromfolder="", tofolder="",
                    use_relativepath=True, usezip=True, filefilter="*.*",
                    directorylevel=1, verbose=0):
        """
        fromfolder:  c:/folder0/folder1/folder2/*
        suffixfolder: /folders1/folder2/
        tofolder: /home/ubuntu/myproject1/
        """
        fromfolder = fromfolder if fromfolder[-1] != '/' else  fromfolder[:-1]
        tmpfolder = '/'.join(fromfolder.split('/')[:-1]) + "/"
        zipname = fromfolder.split('/')[-1] + ".zip"
        print(tmpfolder, zipname)
        zipath = tmpfolder + zipname

        print("zipping")
        filezip = os_zipfolder(dir_tozip=fromfolder, zipname=zipath,
                               dir_prefix=True, iscompress=True)
        print("ssh on remote")
        remote_zipath = os.path.join(tofolder, zipname).replace('\\', '/')
        self.put(filezip, remote_zipath)

        if usezip:
            print("Unzip on remote " + self.host)
            cmd = '/usr/bin/unzip ' + remote_zipath + ' -d ' + tofolder
            ss = self.command(cmd)
            if verbose:
                print(ss)
    if ssh is None and len(host) > 5:
        ssh = aws_ec2_ssh_create_con(contype='ssh', host=host, port=22, username=username,
                                     keyfilepath='')
        print('EC2 connected')

    c = cmdlist
    if isinstance(c, str):  # Only Command to be launched
        if c == 'python': cmfile = AWS().v['AWS_KEYPEM'] if not keyfile else keyfile
        pkey = paramiko.RSAKey.from_private_key_file(keyfile)
        ssh.t.auth_publickey(username, pkey)
        ssh.sftp = paramiko.SFTPClient.from_transport(ssh.t)
        print(command('ls '))


def aws_ec2_ssh_python_script(python_path='/home/ubuntu/anaconda2/bin/ipython',
                              script_path="", args1="", host=""):
   # No space after ipython
    cmd1 = python_path + ' ' + script_path + ' ' + '"' + args1 + '"'
    aws_ec2_ssh_cmd(cmdlist=[cmd1], ssh=None, host=host, username='ubuntu')


def aws_ec2_get_instances(con=None, attributes=None, filters=None, csv_filename=".csv"):
    """
    Fetch all EC2 instances and write selected attributes into a csv file.
    Parameters:
      connection: EC2Connection object
      attributes: Tuple of attributes to retrieve. Default: id, ip_address
      filters={"ip_address":""}
    """
    if not attributes: attributes = AWS().v['EC2_ATTRIBUTES']
    instances = con.get_only_instances(filters=filters)
    instance_list = []
    for instance in instances:
        # attribute is a python keyword.
        x = {attr: getattr(instance, attr) for attr in attributes}
        instance_list.append(x)
    try:
        if len(csv_filename) > 4:
            with open(csv_filename, 'a') as csvfile:
                csvwriter = csv.writer(csvfile, delimiter=",")
                csvwriter.writerow(list(attributes))
                for xi in instance_list:
                    csvwriter.writerow([xi[t] for t in attributes])
            print("EC2 instance", csv_filename)
    except Exception as e:
        print(e)
    return instance_list


def aws_ec2_getfrom_ec2(fromfolder, tofolder, host):
    sftp = aws_ec2_ssh_create_con(contype='sftp', host=host)
    if fromfolder.find('.') > -1:  # file
        folder1, file1 = z_key_splitinto_dir_name(fromfolder[:-1] if fromfolder[-1] == '/' else fromfolder)
        tofolder2 = tofolder if tofolder.find(".") > -1 else  tofolder + '/' + file1
        sftp.get(fromfolder, tofolder2)
    else:  # Pass the Folder in Loop
        pass


def aws_ec2_putfolder(fromfolder='D:/_d20161220/', tofolder='/linux/batch', host=''):
    # fromfolder= DIRCWD +'/linux/batch/task/elvis_prod_20161220/'
    # tofolder=   '/linux/batch/'
    # If you don't care whether the file already exists and you always want to overwrite the
    # files as they are extracted without prompting use the -o switch as follows:
    # https://www.lifewire.com/examples-linux-unzip-command-2201157
    # unzip -o filename.zip
    aws = AWS().v
    folder1, file1 = z_key_splitinto_dir_name(fromfolder[:-1] if fromfolder[-1]=='/' else fromfolder)
    tofolderfull = aws['EC2CWD'] + '/' + tofolder if tofolder.find(aws['EC2CWD']) == -1 else tofolder

    # Zip folder before sending it
    file2 = folder1 + '/' + file1 + '.zip'
    os_zipfolder(fromfolder, file2)
    res = aws_ec2_putfile(file2, tofolder=tofolderfull, host=host)
    print(res)

    # Need install sudo apt-get install zip unzip
    cmd1 = '/usr/bin/unzip '+ tofolderfull + '/' + file1 + '.zip ' + ' -d ' +  tofolderfull + '/'
    aws_ec2_ssh_cmd(cmdlist=[cmd1], host=host)


def aws_ec2_put(fromfolder='d:/file1.zip', tofolder='/home/notebook/aapackage/', host='',
                typecopy='code'):
    """
    Copy python code, copy specific file, copy all folder content
    :param fromfolder: 1 file or 1 folder
    :param tofolder:
    :param host:
    """
    DIRCWD  = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    if fromfolder.find('.') > -1:  # Copy 1 file
        aws_ec2_putfile(fromfolder=fromfolder, tofolder=tofolder, host=host)
    else:  # Copy Folder to
        sftp = aws_ec2_ssh_create_con('sftp', host, isprint=1)
        # Local folder and code folder
        if typecopy == 'code' and fromfolder.find('.') == -1:
            # foldername = fromfolder
            # fromfolder = DIRCWD+ '/' + foldername
            tempfolder = DIRCWD + '/ztemp/' + fromfolder
            # TODO os_folder_delete to be added.
            os_folder_delete(tempfolder)
            os_folder_copy(fromfolder, tempfolder, pattern1="*.py")
            sftp.put(tempfolder, tofolder)
            return 1
        if typecopy == 'all':
            if fromfolder.find(':') == -1:
                print('Please put absolute path')
                return 0
            if fromfolder.find('.') > -1:  #1 file
                fromfolder, file1 = os_split_dir_file(fromfolder)
                tofull = tofolder + '/' + file1 if tofolder.find('.') == -1 else tofolder
                # Not used so commented.
                # tofolder, file2 = util.os_split_dir_file(tofull)
            sftp.put(fromfolder + '/' + file1, tofull)
            try:
                sftp.stat(tofull)
                isexist = True
            except:
                isexist = False
            print(isexist, tofull)


def aws_ec2_putfile(fromfolder='d:/file1.zip', tofolder='/home/notebook/aapackage/',
                    host=''):
    """
    Copy python code, copy specific file, copy all folder content
    :param fromfolder: 1 file or 1 folder
    :param tofolder:
    :param host:
    """
    sftp = aws_ec2_ssh_create_con('sftp', host, isprint=1)
    if fromfolder.find('.') > -1:  # Copy 1 file
        if fromfolder.find(':') == -1:
            print('Please put absolute path')
            return 0
    fromfolder2, file1 = z_key_splitinto_dir_name(fromfolder)
    tofull = tofolder + '/' + file1 if tofolder.find('.') == -1 else tofolder
    print("from: %s, to: %s" % (fromfolder, tofull))
    isexist = False
    try:
        sftp.put(fromfolder, tofull)
        ss = sftp.stat(tofull)
        isexist = True
    except Exception as e:
        print(e)
        ss = None
    sftp.close()
    return isexist, fromfolder, tofull, ss


def sleep2(wsec):
    # from time import sleep
    # from tqdm import tqdm
    for _ in tqdm(range(wsec)):
        sleep(1)


def sftp_isdir(path, sftp):
    try:
        return S_ISDIR(sftp.stat(path).st_mode)
    except IOError:
        #Path does not exist, so by definition not a directory
        return False


def aws_ec2_getfolder(remotepath, sftp):
    paramiko.util.log_to_file('/tmp/paramiko.log')
    def sftp_walk(remotepath):
        path = remotepath
        files = []
        folders = []
        for f in sftp.listdir_attr(remotepath):
            if S_ISDIR(f.st_mode):
                folders.append(f.filename)
            else:
                files.append(f.filename)
        if files:
            yield path, files
        for folder in folders:
            new_path = os.path.join(remotepath, folder)
            for x in sftp_walk(new_path):
                yield x
    for path, files in sftp_walk("." or '/remotepath/'):
        for file1 in files:
            #sftp.get(remote, local) line for dowloading.
            sftp.get(os.path.join(os.path.join(path, file1)), '/local/path/')


############################ UTILS ################################################################
class dict2(object):
    # {} INTO   mydict.key1   ,  mydict.key2 ,   mydict.key4
    def __init__(self, adict):
       self.__dict__.update(adict)


def os_system(cmds, stdout_only=1):
    # Get print output from command line
    cmds = [ t for t  in cmds.split(" ") if t != ""   ]
    try :
      p = subprocess.Popen(cmds, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
      outdata, errdata = p.stdout.read(),  p.stderr.read()
      if stdout_only:
        return outdata
      return outdata, errdata
    except Exception as e:
      print(e)
      return None


def tofloat(value, default=0.0):
    #  Parse the float value.
    try:
        return float(value)
    except:
        return default


# util dependencies
def os_file_getname(path) :
    head, tail = ntpath.split(path)
    return tail or ntpath.basename(head)


def os_file_getpath(path):
    head, _ = ntpath.split(path)
    return head


def os_zipfolder(dir_tozip="", zipname="", dir_prefix=True, iscompress=True):
    """
     shutil.make_archive('/zdisks3/results/output', 'zip', root_dir=/zdisks3/results/',
                         base_dir='output')
     os_zipfolder('zdisk/test/aapackage', 'zdisk/test/aapackage.zip', 'zdisk/test')
     """
    dir_tozip = dir_tozip if dir_tozip[-1] != "/" else dir_tozip[:-1]
    # dir_prefix= dir_prefix if dir_prefix[-1] != '/' else dir_prefix[:-1]

    if dir_prefix:
        dir_tozip, dir_prefix = "/".join(dir_tozip.split("/")[:-1]), dir_tozip.split("/")[-1]
    else:
        dir_tozip, dir_prefix = dir_tozip, "/"

    shutil.make_archive(zipname.replace(".zip", ""), "zip", dir_tozip, base_dir=dir_prefix)
    r = os_zip_checkintegrity(zipname)
    if r:
        return zipname
    else:
        print("Corrupt File")
        return False
    pass

def os_zip_checkintegrity(filezip1):
    zip_file = zipfile.ZipFile(filezip1)
    try:
        ret = zip_file.testzip()
        if ret is not None:
            print("First bad file in zip: %s" % ret)
            return False
        else:
            return True
    except RuntimeError:
        return False


def z_key_splitinto_dir_name(keyname):
    lkey = keyname.split("/")
    if len(lkey) == 1:
        dir1 = ""
    else:
        dir1 = "/".join(lkey[:-1])
        keyname = lkey[-1]
    return dir1, keyname


def os_folder_delete(tempfolder):
    pass

def os_folder_copy(src, dst, symlinks=False, pattern1="*.py", fun_file_toignore=None):
    """
    callable(src, names) -> ignored_names
    'src' parameter, which is the directory being visited by copytree(),
    'names' which is the list of `src` contents, as returned by os.listdir():

    Since copytree() is called recursively, the callable will be called once
     for each directory that is copied.
    It returns a  list of names relative to the `src` directory that should not be copied.
    """

    def _default_fun_file_toignore(src, names):
        _ = src

        pattern = "!" + pattern1
        file_toignore = fnmatch.filter(names, pattern)
        return file_toignore

    if fun_file_toignore is None:
        fun_file_toignore = _default_fun_file_toignore

    try:
        shutil.copytree(src, dst, symlinks=symlinks, ignore=fun_file_toignore)
    except OSError as exc:  # python >2.5
        if exc.errno == errno.ENOTDIR:
            shutil.copy(src, dst, follow_symlinks=False)
        else:
            #raise
            pass


def os_split_dir_file(dirfile):
    lkey = dirfile.split("/")
    if len(lkey) == 1:
        dir1 = ""
    else:
        dir1 = "/".join(lkey[:-1])
        dirfile = lkey[-1]
    return dir1, dirfile


####################################################################################################
############################ CLI ###################################################################
def cli_windows_start_spot():
        # :\_devs\Python01\aws\aapackage\
        # D:\_devs\Python01\ana27\python D:\_devs\Python01\aws\aapackage\util_aws.py --do start_spot
        # aws_ec2_spot_start(EC2_CONN, "west-2", key_name="ecsInstanceRole", inst_type="cx2.2",
        #                    ami_id="", pricemax=0.15,  elastic_ip='',
        #                    pars={"security_group": [""], "disk_size": 25, "disk_type": "ssd",
        #                         "volume_type": "gp2"})
        ss = 'aws ec2 request-spot-instances --region us-west-2 --spot-price "0.55" --instance-count 1 '
        ss += ' --type "one-time" --launch-specification "file://D:\_devs\Python01\\awsdoc\\ec_config2.json" '
        print(ss)
        os.system(ss)
        sleep2(65)
        aws_win_conn = AWS().aws_conn_create_windows()
        ec2_list = aws_ec2_get_instances(aws_win_conn, csv_filename="zz_ec2_instance.csv")
        print(ec2_list)
        for x in ec2_list :
            if x["state"] == "running":
                sleep(5) 
                ss = ' start "Chrome" "C:\Program Files (x86)\Google\Chrome\Application\chrome.exe" '
                ss += ' "http://' +  x["ip_address"] + ':8888/tree#notebooks"  '
                os.system(ss)    


def test_all():
    pass



if __name__ == '__main__':
    print("Start")
    import argparse
    p = argparse.ArgumentParser()  # Command Line input
    p.add_argument('--do', type=str, default='action', help='start_spot')
    p.add_argument('--price', type=float, default=0.5, help='spot price')
    p.add_argument('--spec_file', type=str, default="", help='spec file')
    p.add_argument('--spec_file2', type=str, default="", help='spec file')
    
    args = p.parse_args()
    print(args.do)
    
    if args.do == "test_all":
        test_all()
    
    if args.do == "start_spot_windows":
        cli_windows_start_spot()

    if args.do == "put_file":
        aws_ec2_putfile(fromfolder=args.fromfolder,
                        tofolder=args.tofolder, host=args.host)







# -*- coding: utf-8 -*-
"""Daemon monitoring batch

  ###  monitor subprocess launched in specific folder
       when task is finished, rename folder to _qdone

  ### Folder pattern where process are searched :
    #  /tasks/t53453/main.py   or /tasks/t53453/main.sh  

  ### CLI                                     
batch_daemon_monitor_cli.py --monitor_log_folder   tasks_out/   --monitor_log_file monitor_log_file.log   --log_file   zlog/batchdaemon_monitor.log    --mode daemon 


Folder suffix rules :
 _qstart :  task running
 _qdone  :  task finish
 _ignore :  folder to be ignored
 
 
"""
import argparse
import logging
import os
import sys
from time import sleep

import arrow
import psutil

####################################################################################################
from aapackage import util_log
from aapackage.batch import util_batch, util_cpu

############### Variable definition ################################################################
# MONITOR_LOG_FILE = MONITOR_LOG_FOLDER + "/" + "batch_monitor_" + arrow.utcnow().to('Japan').format("YYYYMMDD_HHmmss") + ".log"
CWD_FOLDER = os.getcwd()


####################################################################################################
logger = logging.basicConfig()


def log(*argv):
    logger.info(",".join([str(x) for x in argv]))


def logcpu(*argv):
    loggercpu.info(",".join([str(x) for x in argv]))


####################################################################################################
def load_arguments():
    parser = argparse.ArgumentParser(description="Monitor python process from tasks")
    parser.add_argument("--verbose", default=0, help="verbose")
    parser.add_argument(
        "--log_file",
        type=str,
        default=CWD_FOLDER + "log_batchdaemon_monitor.log",
        help="daemon log",
    )
    parser.add_argument("--mode", default="nodaemon", help="daemon/ .")

    parser.add_argument(
        "--monitor_log_file",
        type=str,
        default=CWD_FOLDER + "log_batchdaemon_cpu.log",
        help="output the statistics ",
    )
    parser.add_argument(
        "--process_folder", default="/home/ubuntu/tasks/", help="process name pattern"
    )
    parser.add_argument("--process_isregex", default=1, help="process name pattern regex")
    parser.add_argument("--waitsec", type=int, default=10, help="sleep")

    args = parser.parse_args()
    return args


####################################################################################################
if __name__ == "__main__":
    args = load_arguments()
    logger = util_log.logger_setup(
        __name__, log_file=args.log_file, formatter=util_log.FORMATTER_4, isrotate=True
    )

    ### Process CPU usage
    loggercpu = util_log.logger_setup(
        __name__ + "logcpu",
        log_file=args.monitor_log_file,
        formatter=util_log.FORMATTER_4,
        isrotate=True,
    )

    batch_pid_dict = {}
    p_pattern = args.process_folder
    p_pattern = p_pattern[:-1] if p_pattern.endswith("/") else p_pattern

    #  /tasks/t53453/main.py   or /tasks/t53453/main.sh
    regex_pattern = r"((.*/)?%s/t.*/main\.(py|sh))" % p_pattern

    log("Monitor started, regex: ", regex_pattern)
    logcpu("Process monitor started", "", "")
    while True:
        batch_pid = util_cpu.ps_find_procs_by_name(
            name=regex_pattern, ishow=args.verbose, isregex=args.process_isregex
        )

        # Add new PID
        for pid in batch_pid:
            if pid["pid"] not in batch_pid_dict and len(pid["cmdline"]) > 0:
                log("PID added", pid)
                batch_pid_dict[pid["pid"]] = pid
                logcpu("Added", pid["pid"], pid["cmdline"])

        log("PID", batch_pid_dict)

        ddict = {k: v for k, v in batch_pid_dict.items() if v}
        for pid, pr in ddict.items():
            """
         try : 
           status = util_cpu.ps_get_process_status(psutil.Process(pid))
           log(pid, status )         
         except Exception as e :
             log(e)
         """
            try:
                flag = util_cpu.ps_process_isdead(pid)
                if flag:
                    log("Dead", pr)
                    logcpu("Dead", pr["pid"], pr["cmdline"])

                    os.system("pkill -9 " + str(pid))
                    path = pr["cmdline"][1]
                    del batch_pid_dict[pid]

                    path = os.path.dirname(os.path.abspath(path))
                    util_batch.os_folder_rename(path, path.replace("_qstart", "_qdone"))
                    log("_qdone", path)

            except Exception as e:
                log(e)

        if args.mode != "daemon":
            log("Monitor daemon exited")
            break
        sleep(args.waitsec)


"""

if __name__ == '__main__':
    args   = load_arguments()
    logger = util_log.logger_setup(__name__,
                                   log_file  = args.log_file,
                                   formatter = util_log.FORMATTER_4)

    util_batch.os_folder_create(folder= args.monitor_log_folder)

    while True :
      log("Monitor started.")    
      batch_root_pid = util_cpu.ps_find_procs_by_name( name="python",
                                                       cmdline= args.process_pattern )

      for pid in batch_root_pid :
         util_cpu.ps_process_monitor_child( pid,
                                            logfile  = args.monitor_log_file,
                                            duration = args.duration, interval = args.interval)

      log("Monitor Completed.")
      if args.mode != "daemon" : 
        log("Monitor daemon exited")  
        break
      sleep(10)

"""



# -*- coding: utf-8 -*-
# pylint: disable=C0103,W0601,E1123,W0614,F0401,E1120,E1101,E0611,W0702
"""
Launch processors and monitor the CPU, memory usage.
Maintain same level of processors over time.




"""
import csv
import os
import platform
import re
import shlex
import subprocess
import sys
import time
from collections import namedtuple
from datetime import datetime
from time import sleep, time

import arrow
# non-stdlib imports
import psutil

from aapackage import util_log

############# Root folder #####################################################
VERSION = 1


#############Variable #########################################################

######### Logging ##############################################################
LOG_FILE = "zlog/" + util_log.create_logfilename(__file__)
APP_ID = util_log.create_appid(__file__)

logger = util_log.logger_setup(__name__, log_file=None, formatter=util_log.FORMATTER_4)


def log(*argv):
    logger.info(",".join([str(x) for x in argv]))


# log("Ok, test_log")
################################################################################


###############################################################################
def os_getparent(dir0):
    return os.path.abspath(os.path.join(dir0, os.pardir))


# noinspection PyTypeChecker
def ps_process_monitor_child(pid, logfile=None, duration=None, interval=None):
    # We import psutil here so that the module can be imported even if psutil
    # is not present (for example if accessing the version)
    log("Monitoring Started for Process Id: %s" % str(pid))
    pr = psutil.Process(pid)

    # Record start time
    start_time = time.time()

    f = None
    if logfile:
        f = open(logfile, "w")
        f.write(
            "# {0:12s} {1:12s} {2:12s} {3:12s} {4:12s}\n".format(
                "Timestamp".center(12),
                "Elapsed time".center(12),
                "CPU (%)".center(12),
                "Real (MB)".center(12),
                "Virtual (MB)".center(12),
            )
        )

    try:

        # Start main event loop
        while True:

            # Find current time
            current_time = time.time()

            try:
                pr_status = ps_get_process_status(pr)
            except psutil.NoSuchProcess:
                break
            # Check if process status indicates we should exit
            if pr_status in [psutil.STATUS_ZOMBIE, psutil.STATUS_DEAD]:
                log("Process finished ({0:.2f} seconds)".format(current_time - start_time))
                break

            # Check if we have reached the maximum time
            if duration is not None and current_time - start_time > duration:
                break

            # Get current CPU and memory
            try:
                current_cpu = ps_get_cpu_percent(pr)
                # noinspection PyTypeChecker
                current_mem = ps_get_memory_percent(pr)
            except Exception:
                break
            current_mem_real = current_mem.rss / 1024.0 ** 2
            current_mem_virtual = current_mem.vms / 1024.0 ** 2

            # Get information for children
            for child in ps_all_children(pr):
                try:
                    current_cpu += ps_get_cpu_percent(child)
                    current_mem = ps_get_memory_percent(child)
                except Exception:
                    continue
                current_mem_real += current_mem.rss / 1024.0 ** 2
                current_mem_virtual += current_mem.vms / 1024.0 ** 2

            if logfile:
                timestamp = str(arrow.utcnow().to("Japan").format("YYYYMMDD_HHmmss,"))
                f.write(
                    "{0:12} {1:12.3f} {2:12.3f} {3:12.3f} {4:12.3f}\n".format(
                        timestamp,
                        current_time - start_time,
                        current_cpu,
                        current_mem_real,
                        current_mem_virtual,
                    )
                )
                f.flush()

            sleep(interval)

    except KeyboardInterrupt:  # pragma: no cover
        pass

    if logfile:
        f.close()


def ps_wait_process_completion(subprocess_list, waitsec=10):
    for pid in subprocess_list:
        while True:
            try:
                pr = psutil.Process(pid)
                try:
                    pr_status = pr.status()
                except TypeError:  # psutil < 2.0
                    pr_status = pr.status
                except psutil.NoSuchProcess:  # pragma: no cover
                    break
                # Check if process status indicates we should exit
                if pr_status in [psutil.STATUS_ZOMBIE, psutil.STATUS_DEAD]:
                    break
            except:
                break
            time.sleep(waitsec)


def ps_wait_ressourcefree(cpu_max=90, mem_max=90, waitsec=15):
    # wait if computer resources are scarce.
    cpu, mem = 100, 100
    while cpu > cpu_max or mem > mem_max:
        cpu, mem = ps_get_computer_resources_usage()
        time.sleep(waitsec)


def ps_get_cpu_percent(process):
    try:
        return process.cpu_percent()
    except AttributeError:
        return process.get_cpu_percent()


def ps_get_memory_percent(process):
    try:
        return process.memory_info()
    except AttributeError:
        return process.get_memory_info()


def ps_all_children(pr):
    processes = []
    children = []
    try:
        children = pr.children()
    except AttributeError:
        children = pr.get_children()
    except Exception:  # pragma: no cover
        pass

    for child in children:
        processes.append(child)
        processes += ps_all_children(child)
    return processes


def ps_get_process_status(pr):
    try:
        pr_status = pr.status()
    except TypeError:  # psutil < 2.0
        pr_status = pr.status
    except psutil.NoSuchProcess:  # pragma: no cover
        return psutil.STATUS_DEAD
        # raise psutil.NoSuchProcess
    return pr_status


def ps_process_isdead(pid):
    flag = 0
    try:
        pr = psutil.Process(pid)
        pr_status = ps_get_process_status(pr)
        if pr_status in [psutil.STATUS_ZOMBIE, psutil.STATUS_DEAD]:
            flag = 1
    except:
        flag = 1

    return flag


def ps_get_computer_resources_usage():
    cpu_used_percent = psutil.cpu_percent()

    mem_info = dict(psutil.virtual_memory()._asdict())
    # mem_used_percent = 100 - mem_info['available'] / mem_info['total']
    mem_used_percent = mem_info["percent"]

    return cpu_used_percent, mem_used_percent


def ps_find_procs_by_name(name=r"((.*/)?tasks.*/t.*/main\.(py|sh))", ishow=1, isregex=1):
    """ Return a list of processes matching 'name'.
        Regex (./tasks./t./main.(py|sh)|tasks./t.*/main.(py|sh))
        Condensed Regex to:
        ((.*/)?tasks.*/t.*/main\.(py|sh)) - make the characters before 'tasks' optional group.
    """
    ls = []
    for p in psutil.process_iter(["pid", "name", "exe", "cmdline"]):
        cmdline = " ".join(p.info["cmdline"]) if p.info["cmdline"] else ""
        if isregex:
            flag = re.match(name, cmdline, re.I)
        else:
            flag = name and name.lower() in cmdline.lower()

        if flag:
            ls.append({"pid": p.info["pid"], "cmdline": cmdline})

            if ishow > 0:
                log("Monitor", p.pid, cmdline)
    return ls


def os_launch(commands):
    processes = []
    for cmd in commands:
        try:
            p = subprocess.Popen(cmd, shell=False)
            processes.append(p.pid)
            log("Launched: ", p.pid, " ".join(cmd))
            sleep(1)

        except Exception as e:
            log(e)
    return processes


def ps_terminate(processes):
    for p in processes:
        pidi = p.pid
        try:
            os.kill(p.pid, 9)
            log("killed ", pidi)
        except Exception as e:
            log(e)
            try:
                os.kill(pidi, 9)
                log("killed ", pidi)
            except:
                pass


def os_extract_commands(csv_file, has_header=False):
    with open(csv_file, "r", newline="") as file:
        reader = csv.reader(file, skipinitialspace=True)
        if has_header:
            headers = next(reader)  # pass header
        commands = [row for row in reader]

    return commands


def ps_is_issue(p):
    global pars

    pdict = p.as_dict()
    pidi = p.pid

    Mb = 1024 ** 2
    log("Worker PID;CPU;RAM:", pidi, pdict["cpu_percent"], pdict["memory_full_info"][0] / Mb)

    try:
        if not psutil.pid_exists(pidi):
            log("Process has been killed ", pidi)
            return True

        elif pdict["status"] == "zombie":
            log("Process  zombie ", pidi)
            return True

        elif pdict["memory_full_info"][0] >= pars["max_memory"]:
            log("Process  max memory ", pidi)
            return True

        elif pdict["cpu_percent"] >= pars["max_cpu"]:
            log("Process MAX CPU ", pidi)
            return True

        else:
            return False
    except Exception as e:
        log(e)
        return True


def ps_net_send(tperiod=5):
    x0 = psutil.net_io_counters(pernic=False).bytes_sent
    t0 = time()
    sleep(tperiod)
    t1 = time()
    x1 = psutil.net_io_counters(pernic=False).bytes_sent
    return (x1 - x0) / (t1 - t0)


def ps_is_issue_system():
    global net_avg, pars

    try:
        if psutil.cpu_percent(interval=5) > pars["cpu_usage_total"]:
            return True

        elif psutil.virtual_memory().available < pars["mem_available_total"]:
            return True

        else:
            return False

    except:
        return True


def monitor_maintain():
    """
       Launch processors and monitor the CPU, memory usage.
       Maintain same leve of processors over time.
    """
    global pars

    log("start monitoring", len(CMDS))
    cmds2 = []
    for cmd in CMDS:
        ss = shlex.split(cmd)
        cmds2.append(ss)

    processes = os_launch(cmds2)
    try:
        while True:
            has_issue = []
            ok_process = []
            log("N_process", len(processes))

            ### check global system  ##########################################
            if len(processes) == 0 or ps_is_issue_system():
                log("Reset all process")
                lpp = ps_find_procs_by_name(pars["proc_name"], 1)
                ps_terminate(lpp)
                processes = os_launch(cmds2)
                sleep(5)

            ## pid in process   ###############################################
            for pidi in processes:
                try:
                    p = psutil.Process(pidi)
                    log("Checking", p.pid)

                    if ps_is_issue(p):
                        has_issue.append(p)

                    else:
                        log("Process Fine ", pidi)
                        ok_process.append(p)

                except Exception as e:
                    log(e)

            ### Process with issues    ########################################
            for p in has_issue:
                try:
                    log("Relaunching", p.pid)
                    pcmdline = p.cmdline()
                    pidlist = os_launch([pcmdline])  # New process can start before

                    sleep(3)
                    ps_terminate([p])
                except:
                    pass

            ##### Check the number of  processes    ###########################
            sleep(5)
            lpp = ps_find_procs_by_name(pars["proc_name"], 1)

            log("Active process", len(lpp))
            if len(lpp) < pars["nproc"]:
                for i in range(0, pars["nproc"] - len(lpp)):
                    pidlist = os_launch([shlex.split(pars["proc_cmd"])])

            else:
                for i in range(0, len(lpp) - pars["nproc"]):
                    pidlist = ps_terminate([lpp[i]])

            sleep(5)
            lpp = ps_find_procs_by_name(pars["proc_name"], 0)
            processes = [x["pid"] for x in lpp]

            log("Waiting....")
            sleep(1)

    except Exception as e:
        log(e)


############ AZURE NODE #################################################################
#########################################################################################
"""
[  arg.logfolder, arg.name, arg.consumergroup, arg.input_topic,  
   arrow.utcnow().to('Japan').format("YYYYMMDD_HHmm_ss"),
     str(random.randrange(1000))])
# DIRCWD = os_getparent(os.path.dirname(os.path.abspath(__file__)))

"""
_DEFAULT_STATS_UPDATE_INTERVAL = 5
_OS_DISK = None
_USER_DISK = None
"""
_IS_PLATFORM_WINDOWS = platform.system() == 'Windows'
if _IS_PLATFORM_WINDOWS:
    _OS_DISK = 'C:/' # This is inverted on Cloud service
    _USER_DISK = 'D:/'
else:
    _OS_DISK = "/"
    _USER_DISK = '/mnt/resources'
    if not os.path.exists(_USER_DISK):
        _USER_DISK = '/mnt'
"""


def os_python_environment():  # pragma: no cover
    return " ".join([platform.python_implementation(), platform.python_version()])


def os_environment():
    return platform.platform()


def os_is_wndows():
    return platform.system() == "Windows"


def np_avg(list):
    return sum(list) / float(len(list))


def np_pretty_nb(num, suffix=""):
    for unit in ["", "K", "M", "G", "T", "P", "E", "Z"]:
        if abs(num) < 1000.0:
            return "%3.1f%s%s" % (num, unit, suffix)
        num /= 1000.0
    return "%.1f%s%s" % (num, "Yi", suffix)


NodeIOStats = namedtuple("NodeIOStats", ["read_bps", "write_bps"])


class NodeStats:
    """Persistent Task Stats class"""

    def __init__(
        self,
        num_connected_users=0,
        num_pids=0,
        cpu_count=0,
        cpu_percent=None,
        mem_total=0,
        mem_avail=0,
        swap_total=0,
        swap_avail=0,
        disk_io=None,
        disk_usage=None,
        net=None,
    ):
        """
        Map the attributes
        """
        self.num_connected_users = num_connected_users
        self.num_pids = num_pids
        self.cpu_count = cpu_count
        self.cpu_percent = cpu_percent
        self.mem_total = mem_total
        self.mem_avail = mem_avail
        self.swap_total = swap_total
        self.swap_avail = swap_avail
        self.disk_io = disk_io or NodeIOStats(0, 0)
        self.disk_usage = disk_usage or dict()
        self.net = net or NodeIOStats(0, 0)

    @property
    def mem_used(self):
        """
            Return the memory used
        """
        return self.mem_total - self.mem_avail


class IOThroughputAggregator:
    def __init__(self):
        self.last_timestamp = None
        self.last_read = 0
        self.last_write = 0

    def aggregate(self, cur_read, cur_write):
        """
            Aggregate with the new values
        """
        now = datetime.now()
        read_bps = 0
        write_bps = 0
        if self.last_timestamp:
            delta = (now - self.last_timestamp).total_seconds()
            read_bps = (cur_read - self.last_read) / delta
            write_bps = (cur_write - self.last_write) / delta

        self.last_timestamp = now
        self.last_read = cur_read
        self.last_write = cur_write

        return NodeIOStats(read_bps, write_bps)


class NodeStatsCollector:
    """
    Node Stats Manager class
    """

    def __init__(
        self,
        pool_id,
        node_id,
        refresh_interval=_DEFAULT_STATS_UPDATE_INTERVAL,
        app_insights_key=None,
    ):
        self.pool_id = pool_id
        self.node_id = node_id
        self.telemetry_client = None
        self.first_collect = True
        self.refresh_interval = refresh_interval

        self.disk = IOThroughputAggregator()
        self.network = IOThroughputAggregator()

        if (
            app_insights_key
            or "APP_INSIGHTS_INSTRUMENTATION_KEY" in os.environ
            or "APP_INSIGHTS_KEY" in os.environ
        ):
            key = (
                app_insights_key
                or os.environ.get("APP_INSIGHTS_INSTRUMENTATION_KEY")
                or os.environ.get("APP_INSIGHTS_KEY")
            )

            # log("Detected instrumentation key. Will upload stats to app insights")
            # self.telemetry_client = TelemetryClient(key)
            # context = self.telemetry_client.context
            # context.application.id = "AzureBatchInsights"
            # context.application.ver = VERSION
            # context.device.model = "BatchNode"
            # context.device.role_name = self.pool_id
            # context.device.role_instance = self.node_id
        else:
            log(
                "No instrumentation key detected. Cannot upload to app insights."
                + "Make sure you have the APP_INSIGHTS_INSTRUMENTATION_KEY environment variable setup"
            )

    def init(self):
        """
            Initialize the monitoring
        """
        # start cpu utilization monitoring, first value is ignored
        psutil.cpu_percent(interval=None, percpu=True)

    def _get_network_usage(self):
        netio = psutil.net_io_counters()
        return self.network.aggregate(netio.bytes_recv, netio.bytes_sent)

    def _get_disk_io(self):
        diskio = psutil.disk_io_counters()
        return self.disk.aggregate(diskio.read_bytes, diskio.write_bytes)

    def _get_disk_usage(self):
        disk_usage = dict()
        try:
            disk_usage[_OS_DISK] = psutil.disk_usage(_OS_DISK)
            disk_usage[_USER_DISK] = psutil.disk_usage(_USER_DISK)
        except Exception as e:
            logger.error("Could not retrieve user disk stats for {0}: {1}".format(_USER_DISK, e))
        return disk_usage

    def _sample_stats(self):
        # get system-wide counters
        mem = psutil.virtual_memory()
        disk_stats = self._get_disk_io()
        disk_usage = self._get_disk_usage()
        net_stats = self._get_network_usage()

        swap_total, _, swap_avail, _, _, _ = psutil.swap_memory()

        stats = NodeStats(
            cpu_count=psutil.cpu_count(),
            cpu_percent=psutil.cpu_percent(interval=None, percpu=True),
            num_pids=len(psutil.pids()),
            # Memory
            mem_total=mem.total,
            mem_avail=mem.available,
            swap_total=swap_total,
            swap_avail=swap_avail,
            # Disk IO
            disk_io=disk_stats,
            # Disk usage
            disk_usage=disk_usage,
            # Net transfer
            net=net_stats,
        )
        del mem
        return stats

    def _collect_stats(self):
        """
            Collect the stats and then send to app insights
        """
        # collect stats
        stats = self._sample_stats()

        if self.first_collect:
            self.first_collect = False
            return

        if stats is None:
            logger.error("Could not sample node stats")
            return

        if self.telemetry_client:
            self._send_stats(stats)
        else:
            self._log_stats(stats)

    def _send_stats(self, stats):
        """
            Retrieve the current stats and send to app insights
        """
        process = psutil.Process(os.getpid())

        logger.debug(
            "Uploading stats. Mem of this script: %d vs total: %d",
            process.memory_info().rss,
            stats.mem_avail,
        )
        client = self.telemetry_client

        for cpu_n in range(0, stats.cpu_count):
            client.track_metric("Cpu usage", stats.cpu_percent[cpu_n], properties={"Cpu #": cpu_n})

        for name, disk_usage in stats.disk_usage.items():
            client.track_metric("Disk usage", disk_usage.used, properties={"Disk": name})
            client.track_metric("Disk free", disk_usage.free, properties={"Disk": name})

        client.track_metric("Memory used", stats.mem_used)
        client.track_metric("Memory available", stats.mem_avail)
        client.track_metric("Disk read", stats.disk_io.read_bps)
        client.track_metric("Disk write", stats.disk_io.write_bps)
        client.track_metric("Network read", stats.net.read_bps)
        client.track_metric("Network write", stats.net.write_bps)
        self.telemetry_client.flush()

    def _log_stats(self, stats):
        log("========================= Stats =========================")
        log("Cpu percent:            %d%% %s", np_avg(stats.cpu_percent), stats.cpu_percent)
        log(
            "Memory used:       %sB / %sB",
            np_pretty_nb(stats.mem_used),
            np_pretty_nb(stats.mem_total),
        )
        log(
            "Swap used:         %sB / %sB",
            np_pretty_nb(stats.swap_avail),
            np_pretty_nb(stats.swap_total),
        )
        log("Net read:               %sBs", np_pretty_nb(stats.net.read_bps))
        log("Net write:              %sBs", np_pretty_nb(stats.net.write_bps))
        log("Disk read:               %sBs", np_pretty_nb(stats.disk_io.read_bps))
        log("Disk write:              %sBs", np_pretty_nb(stats.disk_io.write_bps))
        log("Disk usage:")
        for name, disk_usage in stats.disk_usage.items():
            log("  - %s: %i/%i (%i%%)", name, disk_usage.used, disk_usage.total, disk_usage.percent)

        log("-------------------------------------")
        log("")

    def run(self):
        """
            Start collecting information of the system.
        """
        logger.debug("Start collecting stats for pool=%s node=%s", self.pool_id, self.node_id)
        while True:
            self._collect_stats()
            time.sleep(self.refresh_interval)


def monitor_nodes():
    """
    Main entry point for prism
    """
    # log basic info
    log("Python args: %s", sys.argv)
    log("Operating system: %s", os_environment())
    log("Cpu count: %s", psutil.cpu_count())

    pool_id = os.environ.get("AZ_BATCH_POOL_ID", "_test-pool-1")
    node_id = os.environ.get("AZ_BATCH_NODE_ID", "_test-node-1")

    # get and set event loop mode
    log("enabling event loop debug mode")

    app_insights_key = None
    if len(sys.argv) > 2:
        pool_id = sys.argv[1]
        node_id = sys.argv[2]
    if len(sys.argv) > 3:
        app_insights_key = sys.argv[3]

    # create node stats manager
    collector = NodeStatsCollector(pool_id, node_id, app_insights_key=app_insights_key)
    collector.init()
    collector.run()


def os_generate_cmdline():
    Mb = 1024 ** 2
    pars = {
        "max_memory": 1500.0 * Mb,
        "max_cpu": 85.0,
        "proc_name": "streaming_couchbase_update_cli.py",
        "mem_available_total": 2000.0 * Mb,
        "cpu_usage_total": 98.0,
    }
    CMDS = [pars["proc_cmd"]] * pars["nproc"]
    return CMDS, pars


if __name__ == "__main__":
    ################## Initialization #########################################
    log(" Log check")

    CMDS, pars = os_generate_cmdline()

    ############## RUN Monitor ################################################
    # monitor()



# -*- coding: utf-8 -*-
"""
Batch utils
    main.sh
    main.py



"""
import os
import random
import subprocess
import sys
import time

import numpy as np
import pandas as pd

from aapackage import util_log
#######################################################################

################### Variables #########################################


######### Logging ####################################################
global logger
LOG_FILE = "zlog/" + util_log.create_logfilename(__file__)
APP_ID = util_log.create_appid(__file__)

# noinspection PyRedeclaration
logger = util_log.logger_setup(__name__, log_file=None, formatter=util_log.FORMATTER_4)


def log(*argv):
    logger.info(",".join([str(x) for x in argv]))


# log("Ok, test_log")
#####################################################################


######################################################################
def os_python_path():
    return str(sys.executable)


def os_folder_rename(old_folder, new_folder):
    try:
        if os.path.isdir(new_folder):
            new_folder = new_folder + str(random.randint(100, 999))
        os.rename(old_folder, new_folder)
        return new_folder
    except Exception as e:
        return old_folder


def os_folder_create(folder):
    if not os.path.isdir(folder):
        os.makedirs(folder)


def os_cmd_generate(task_folder, os_python_path=None):
    """
   params.toml check
   otherwise default config is/main.py with current interpreter.
     main.sh 
     main.py
     
  """
    main_file = os.path.join(task_folder, "main.sh")
    if os.path.isfile(main_file):
        cmd = [main_file]
        return cmd

    main_file = os.path.join(task_folder, "main.py")
    if os.path.isfile(main_file):
        os_python_path = sys.executable if os_python_path is None else os_python_path
        cmd = [os_python_path, main_file]
        return cmd


def os_wait_policy(waitsleep=15, cpu_max=95, mem_max=90.0):
    """
      Wait when CPU/Mem  usage is too high 
    
    """
    from aapackage.batch import util_cpu

    cpu_pct, mem_pct = util_cpu.ps_get_computer_resources_usage()
    while cpu_pct > cpu_max or mem_pct > mem_max:
        log("cpu,mem usage", cpu_pct, mem_pct)
        cpu_pct, mem_pct = util_cpu.ps_get_computer_resources_usage()
        time.sleep(waitsleep)


def batch_run_infolder(
    task_folders,
    suffix="_qstart",
    main_file_run="main.py",
    waitime=7,
    os_python_path=None,
    log_file=None,
):
    sub_process_list = []
    global logger

    if ".py" in main_file_run:
        ispython = 1

    if os_python_path is None:
        os_python_path = sys.executable

    if log_file is not None:
        logger = util_log.logger_setup(__name__, log_file=log_file, formatter=util_log.FORMATTER_4)

    for folder_i in task_folders:
        foldername = folder_i + suffix
        foldername = os_folder_rename(old_folder=folder_i, new_folder=foldername)

        # main_file = os.path.join(foldername,  main_file_run )
        # cmd = [os_python_path, main_file]  if ispython else [ main_file]
        cmd = os_cmd_generate(foldername, os_python_path)

        os_wait_policy(waitsleep=15)
        ps = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell=False)
        sub_process_list.append(ps.pid)

        log("Sub-process, ", ps.pid, cmd)
        time.sleep(waitime)

    return sub_process_list


def batch_parallel_subprocess(hyperparam_file, subprocess_script, os_python_path=None, waitime=5):
    """
    task/
          main.py
          subprocess.py
          hyperparams.csv
          ...
          :type hyperparam_file: str

    """
    hyper_parameter = pd.read_csv(hyperparam_file)
    PYTHON_PATH = sys.executable if os_python_path is None else os_python_path
    ispython = 1 if ".py" in subprocess_script else 0

    # Start process launch
    subprocess_list = []
    for ii in range(0, len(hyper_parameter)):
        if ispython:
            cmd = [PYTHON_PATH, subprocess_script, "--hyperparam_ii=%d" % ii]
        else:
            cmd = [subprocess_script, ii]

        ps = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell=False)
        log("Subprocess started ", ps.pid)
        subprocess_list.append(ps.pid)
        time.sleep(waitime)
        # util_cpu.ps_wait_ressourcefree(cpu_max=90, mem_max=90, waitsec=15)

    # util_cpu.ps_wait_process_completion(subprocess_list)


def batch_generate_hyperparameters(hyperparam_dict, outfile_hyperparam):
    """
     {  "param1" : {"min": 10  , "max": 200 , "type": int,    "nmax": 10, "method": "random"  },
        "param2" : {"min": 10  , "max": 50 , "type": float,  "nmax": 10  }, "method": "linear"
     }

    df
      nbcols    :  range for param1  X  range for param2  X range for param3
      nbcolumns :  len(hyper_dict)

   https://stackoverflow.com/questions/42795832/faster-way-to-extend-values-of-a-pandas-dataframe

   Dict preserving the order !!!

    Equivalent of grid search
https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.GridSearchCV.html


  """
    # df not defined , DataFrame has no "extend" method
    ##init columns

    df = pd.DataFrame({k for k in hyperparam_dict.keys()})

    for key, d in hyperparam_dict.items():
        df[key] = 0

        vv = np.arange(d["min"], d["max"], d["nmax"])

        # df = df.extend(  len(vv) )

    # df.to_csv( file_hyper )
    return 1






if __name__ == "__main__":
    ################## Initialization #########################################
    log(" Log check")

    ############## RUN Monitor ################################################
    # monitor()


"""
def checkAdditionalFiles(WorkingFolder, AdditionalFiles):
    if not AdditionalFiles:
        return []

    AdditionalFiles = [f.strip(" ")
                       for f in AdditionalFiles.split(",")]

    for File in AdditionalFiles:
        ExpectedFilePath = os.path.isfile(os.path.join(WorkingFolder, File))
        if not ExpectedFilePath:
            print("Additional file <%s> %s not found. Aborting!" % File)
            exit(1)

    return AdditionalFiles


def os_create_folder(WorkingFolder, folderName):
    folderPath = os.path.join(WorkingFolder, folderName)
    if not os.path.isdir(folderPath):
        os.mkdir(folderPath)





"""


"""
date0 = util.date_now()
isverbose = 0
batchname = task_folder
DIRBATCH = DIRCWD + '/linux/batch/task/' + task_folder + '/'

batch_script = DIRBATCH + '/elvis_pygmo_optim.py'
batch_in_data1 = DIRBATCH + '/aafolio_storage_ref.pkl'
filedata = DIRBATCH + '/close_15assets_2012_dec216.spydata'


dir_batch_main = DIRCWD + '/linux/batch/'
# str(np.random.randint(1000, 99999999))
batch_out_name = 'batch_' + util.date_nowtime('stamp')
batch_out = dir_batch_main + '/output/'+date0 + '/' + batch_out_name
os.makedirs(batch_out)

batch_out_log = batch_out + '/output_result.txt'
batch_out_data = batch_out + '/aafolio_storage_' + date0 + '.pkl'
util.os_print_tofile('\n\n'+title1, batch_out_log)
"""


"""
if util.os_file_exist(batch_out_data):
    aux3_cols, aafolio_storage = util.py_load_obj(
        batch_out_data, isabsolutpath=1)
else:
    aux3_cols, aafolio_storage = util.py_load_obj(
        batch_in_data1, isabsolutpath=1)


"""


"""
        No Need t
        # build path & create task directory;
        TaskFolderName = "task_%s_%i" % (batch_label, ii)
        TaskFolder = os.path.join(WorkingFolder, TaskFolderName)

        os.mkdir(TaskFolder)

        # Copy optimization script to Task Folder;
        shutil.copy(os.path.join(WorkingFolder, OptimizerName),
                    os.path.join(TaskFolder, OptimizerName))

        # Save Parameters inside Task Folder;
        with open(os.path.join(TaskFolder, "parameters.toml"), "w") as task_parameters:
            toml.dump(params_dict, task_parameters)

        # Copy additional files to Task Folder;
        if AdditionalFilesArg:
            AdditionalFiles = checkAdditionalFiles(WorkingFolder, AdditionalFilesArg)
            for File in AdditionalFiles:
                shutil.copy(os.path.join(WorkingFolder, File),
                            os.path.join(TaskFolder, File))
"""



# -*- coding: utf-8 -*-
"""
  Daemon monitoring batch
  scans sub-folder in /tasks/
  and execute  /tasks/taskXXXX/main.py in that folder  (if task is not qdone)
                               main.sh
  When task is finished, taskXXX is rename to taskXXX_qdone
_qstart
_qdone  : qdone
_ignore : ignore file
### CLI usage
batch_daemon_launch_cli.py  --task_folder tasks  --log_file zlog/batchdaemong.log  --mode daemon  --waitsec 10  &
### using S3 disk
batch_daemon_launch_cli.py  --task_folder  zs3drive/tasks  --log_file   zlog/batchdaemong.log  --mode daemon  --waitsec 10  
batch_daemon_monitor_cli.py --monitor_log_folder   tasks_out/   --monitor_log_file monitor_log_file.log   --log_file   zlog/batchdaemon_monitor.log    --mode daemon     
########## Details 
batch_daemon_launch_cli.py --param_file zs3drive/config_batch.toml --param_mode test_launch
###############################################################################
cp zs3drive/tasks/ztask_test1_ignore   zs3drive/tasks/task_test1 --recursive
rm zs3drive/tasks/ztask_test1_ignore  --recursive
"""
import argparse
import json
import logging
import os
import subprocess
import sys
import time
from time import sleep

################################################################################
from aapackage import util_log
from aapackage.batch import util_batch, util_cpu

############### logger #########################################################
# DIR_PATH = os.path.dirname(os.path.realpath(__file__))
# TASK_FOLDER_DEFAULT = os.getcwd()
TASK_FOLDER_DEFAULT = os.path.dirname(os.path.realpath(__file__)) + "/ztestasks/"


################################################################################
global logger
logger = logging.basicConfig()


def log(*argv):
    logger.info(",".join([str(x) for x in argv]))


################################################################################
def load_arguments():
    """
       --param_file /zs3drive/config_batch.toml --param_mode test_launch
  """
    cur_path = os.path.dirname(os.path.realpath(__file__))
    config_file = os.path.join(cur_path, "config.toml")

    p = argparse.ArgumentParser()
    p.add_argument("--param_file", default=config_file, help="Params File")
    p.add_argument("--param_mode", default="test", help=" test/ prod /uat")

    p.add_argument("--task_folder", help="path to task folder.")  #     default=TASK_FOLDER_DEFAULT
    p.add_argument("--log_file", help=".")  # default="logfile_batchdaemon.log"
    p.add_argument("--log_file_task", help=".")  #     default="logfile_batchdaemon_task.log",
    p.add_argument("--mode", help="daemon/ .")  #     default="nodaemon",
    p.add_argument("--waitsec", type=int, help="wait sec")  #  default=30,
    p.add_argument(
        "--global_task_file", help="synchronize task"
    )  # default="/home/ubuntu/zs3drive/global_task.json",

    args = p.parse_args()

    ##### Load default file params as dict namespace #########################
    import toml

    class to_namespace(object):
        def __init__(self, adict):
            self.__dict__.update(adict)

    try:
        pars = toml.load(args.param_file)  # Load Default params
        # print(pars)
        pars = pars[args.param_mode]  # test / prod
        # print(pars)

        ### Overwrite params by CLI input and merge with toml file
        for key, x in vars(args).items():
            if x is not None:  # only values NOT set by CLI
                pars[key] = x

        print(pars)
        pars = to_namespace(pars)  #  like object/namespace pars.instance
        return pars
    except Exception as e:
        print("Load_argument error", e)
        return args


################################################################################
################################################################################
def get_list_valid_task_folder(folder, script_name="main"):
    if not os.path.isdir(folder):
        return []

    valid_folders = []
    for root, dirs, files in os.walk(folder):
        root_splits = root.split("/")
        for filename in files:
            if (
                filename == "main.sh"
                or filename == "main.py" "_qstart" not in root_splits[-1]
                and "_qdone" not in root_splits[-1]
                and "_ignore" not in root_splits[-1]
            ):
                valid_folders.append(root)

    return valid_folders


def subprocess_launch(foldername, main_file):
    if main_file == "main.py":
        os_python_path = sys.executable
        cmd = [os_python_path, os.path.abspath(foldername + "/" + main_file)]
    else:
        main_file = os.path.abspath(foldername + "/" + main_file)
        os.system("chmod 777 " + main_file)
        cmd = ["bash", main_file]  # main.sh
        # cmd = [ "chmod 777 " + main_file  + " && " + main_file]  # main.sh

    ps = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell=False)
    return ps.pid


def os_wait_policy(waitsleep=15, cpu_max=95, mem_max=90.0):
    """
      Wait when CPU/Mem  usage is too high 
    """
    from aapackage.batch import util_cpu

    cpu_pct, mem_pct = util_cpu.ps_get_computer_resources_usage()
    log("cpu,mem usage", cpu_pct, mem_pct)
    while cpu_pct > cpu_max or mem_pct > mem_max:
        log("cpu,mem usage", cpu_pct, mem_pct)
        cpu_pct, mem_pct = util_cpu.ps_get_computer_resources_usage()
        time.sleep(waitsleep)


####################################################################################################
####################################################################################################
def isvalid_folder(folder_main, folder, folder_check, global_task_file):
    if os.path.isfile(os.path.join(folder_main, folder)) or folder in folder_check:
        return False
    elif "_qdone" in folder or "_qstart" in folder or "_ignore" in folder:
        global_task_file_save(folder, folder_check, global_task_file)
        return False
    else:
        return True


def global_task_file_save(folder, folder_check, global_task_file):
    folder_check[folder] = time.time()
    json.dump(folder_check, open(global_task_file, mode="w"))
    log("Inserted task", folder, global_task_file)


def main3():
    """ Driver utility for the script.
    global_task_file contains the list of task ALREADY PROCESSED.
    if a task is pick up --> this file is updated Globally.
  
  """
    global logger
    args = load_arguments()
    logger = util_log.logger_setup(
        __name__, log_file=args.log_file, formatter=util_log.FORMATTER_4, isrotate=True
    )

    log("Daemon", "start ", os.getpid())
    folder_main = args.task_folder
    global_task_file = args.global_task_file
    if not os.path.isdir(folder_main):
        return 0

    folder_check = json.load(open(global_task_file, mode="r"))
    while True:
        # log(folder_check)
        log("Daemon new loop", folder_main)
        for folder in os.listdir(folder_main):
            # log(folder)
            if not isvalid_folder(folder_main, folder, folder_check, global_task_file):
                continue

            t0 = time.time()
            folder_check = json.load(open(global_task_file, mode="r"))  # Refresh Global file
            if folder not in folder_check:
                global_task_file_save(
                    folder, folder_check, global_task_file
                )  # Update to prevent 2nd pick up
                log("time to save", time.time() - t0)

                folder = os.path.join(folder_main, folder)
                files = [
                    file for file in os.listdir(folder) if file == "main.sh" or file == "main.py"
                ]
                log(files)
                if files:
                    pid = subprocess_launch(folder, files[0])
                    log("task folder started:", folder, files[0], pid)
                    sleep(20)
                    os_wait_policy(waitsleep=10)

        if args.mode != "daemon":
            log("Daemon", "terminated", os.getpid())
            break

        sleep(args.waitsec)
        os_wait_policy(waitsleep=5)


if __name__ == "__main__":
    main3()


####################################################################################################
####################################################################################################


"""
class global_task(object):
    def __init__(self, task_file):
       self.__dict__.update(adict)
  
    def save(self) :
      pass
      
    def load(self):
      pass
"""


def main():
    """ Driver utility for the script."""
    global logger
    args = load_arguments()
    logger = util_log.logger_setup(
        __name__, log_file=args.log_file, formatter=util_log.FORMATTER_4, isrotate=True
    )

    log("Daemon", "start ", os.getpid())
    while True:
        log("Daemon new loop", args.task_folder)
        folders = get_list_valid_task_folder(args.task_folder)

        if folders:
            log("task folder:", folders)
            pid_list = util_batch.batch_run_infolder(task_folders=folders, log_file=args.log_file)
            log("task folder started:", pid_list)

        if args.mode != "daemon":
            log("Daemon", "terminated", os.getpid())
            break

        sleep(args.waitsec)


def main2():
    """ Driver utility for the script."""
    global logger
    args = load_arguments()
    logger = util_log.logger_setup(
        __name__, log_file=args.log_file, formatter=util_log.FORMATTER_4, isrotate=True
    )

    log("Daemon", "start ", os.getpid())
    folder_main = args.task_folder
    while True:
        log("Daemon new loop", folder_main)
        if not os.path.isdir(folder_main):
            break

        for root, dirs, files in os.walk(folder_main):
            root_splits = root.split("/")
            f = root_splits[-1]
            for filename in files:
                if (
                    filename == "main.sh"
                    or filename == "main.py"
                    and "_qstart" not in f
                    and "_qdone" not in f
                    and "_ignore" not in f
                ):
                    try:
                        #### Issue of collision if 2 instances rename the folder
                        folder_new = root + "_qstart"
                        os.rename(root, folder_new)

                        pid = subprocess_launch(folder_new, filename)
                        log("task folder started:", folder_new, pid)
                    except:
                        pass
                    os_wait_policy(waitsleep=5)

        if args.mode != "daemon":
            log("Daemon", "terminated", os.getpid())
            break

        sleep(args.waitsec)
        os_wait_policy(waitsleep=5)


"""
################### Argument catching ########################################################
print('Start Args')
xprint sys.argv, "\n\n", sys.argv[0], "\n\n", sys.argv[1], "\n\n"
title1= sys.argv[0]
task_folder= title1.split('/')[-2]            #  print 'Task Folder', task_folder
#Task data, str  --> Tuple of arguments
args1=  eval(sys.argv[1])
input_param_file, itask0, itask1= args1[0]
#Other data
args2= args1[1]
print(title1, input_param_file, (itask0, itask1), args2 )
################### Batch Params #############################################################
date0= util.date_now()
isverbose =0
batchname=        task_folder
DIRBATCH=         DIRCWD + '/linux/batch/task/'+ task_folder + '/'
batch_script=     DIRBATCH + '/elvis_pygmo_optim.py'
batch_in_data1=   DIRBATCH + '/aafolio_storage_ref.pkl'
filedata=         DIRBATCH + '/close_15assets_2012_dec216.spydata'
dir_batch_main=   DIRCWD + '/linux/batch/'
batch_out_name=   'batch_' + util.date_nowtime('stamp')   #str(np.random.randint(1000, 99999999))
batch_out=        dir_batch_main + '/output/'+date0 + '/'+ batch_out_name
os.makedirs(batch_out)
batch_out_log=    batch_out + '/output_result.txt'
batch_out_data=   batch_out + '/aafolio_storage_' + date0 + '.pkl'
util.os_print_tofile( '\n\n'+title1, batch_out_log)
################  Output Data table ######################################################
if util.os_file_exist(batch_out_data):
  aux3_cols, aafolio_storage=  util.py_load_obj( batch_out_data, isabsolutpath=1 )
else :
  aux3_cols, aafolio_storage=  util.py_load_obj( batch_in_data1, isabsolutpath=1 )
################## Model Parameters ######################################################
# input_param_file=    DIRBATCH_local+'input_params_all.pkl'
input_params_list=   util.py_load_obj(input_param_file, isabsolutpath=1)
input_params_list=   input_params_list[itask0:itask1,:]    #Reduce the size
############### Loop on each parameters sets #############################################
iibatch=0; krepeat= 1;ii=0
for ii in xrange(itask0-itask0, itask1-itask0):
  iibatch+=1
  params_dict= dict(input_params_list[ii,1])
  globals().update(params_dict)     #Update in the interpreter memory
  util.os_print_tofile('\n ' + task_folder + ' , task_' + str(ii+itask0) + '\n', batch_out_log)
  for kkk in xrange(0, krepeat) :   # Random Start loop  krepeat to have many random start....
      execfile(batch_script)
"""

"""  Manual test
  ii= 2; kkk=1
  params_dict= dict(input_params_list[ii,1])
  globals().update(params_dict)
  util.os_print_tofile('\n task_' + str(ii) +'\n' , batch_out_log)
  for kkk in xrange(0, krepeat) : # Random Start loop
      execfile(batch_script)
"""

#####################################################################################

""" Create Storage file :
np.concatenate((    ))
aafolioref= aafolio_storage[0,:].reshape(1,20)
util.py_save_obj( (aux3_cols, aafolioref) ,  'aafolio_storage_ref' )
aux3_cols, aafolio_storage=  util.py_load_obj( dir_batch_main+  '/batch_20161228_96627894/aafolio_storage_20161227',
                                                   isabsolutpath=1 )
"""

#####################################################################################

"""
for arg in sys.argv:
    print arg
Each command-line argument passed to the program will be in sys.argv, which is just a list. Here you are printing each argument on a separate line.
Example 10.21. The contents of sys.argv
[you@localhost py]$ python argecho.py             1
argecho.py
"""

"""
import argparse
p.add_argument('-i','--input', help='Script File Name', required=False)
p.add_argument('-o','--output',help='Script ID', required=False)
args = p.parse_args()
## show values ##
print ("Input file: %s" % args.input )
print ("Output file: %s" % args.output )
"""



"""
from . import batch_sequencer
from functools import partial

options = batch_sequencer.load_arguments()

mandatoryArguments = (
    options.HyperParametersFile,
    options.WorkingDirectory,
    options.OptimizerName,
    options.AdditionalFiles
)


build_execute_batch = batch_sequencer.build_execute_batch

execute_batch = partial(
    build_execute_batch,
    *mandatoryArguments)

"""


import logging
from logging import NullHandler

logging.getLogger(__name__).addHandler(NullHandler())



# -*- coding: utf-8 -*-
"""
https://docs.python-guide.org/writing/logging/
https://docs.python.org/3/howto/logging-cookbook.html


my_logger = util_log.logger_setup("my module name", log_file="")
APP_ID    = util_log.create_appid(__file__ )
def log(s1='', s2='', s3='', s4='', s5='', s6='', s7='', s8='', s9='', s10='') :
       my_logger.debug( ",".join( [APP_ID, str(s1), str(s2), str(s3), str(s4), str(s5) ,
                        str(s6), str(s7), str(s8), str(s9), str(s10)] ) )


Attribute name	Format	Description
args	You shouldn’t need to format this yourself.	The tuple of arguments merged into msg to
produce message, or a dict whose values are used for the merge (when there is only one argument,
and it is a dictionary).
asctime	%(asctime)s	Human-readable time when the LogRecord was created. By default this is of
the form ‘2003-07-08 16:49:45,896’ (the numbers after the comma are millisecond portion of the
time).
created	%(created)f	Time when the LogRecord was created (as returned by time.time()).
exc_info	You shouldn’t need to format this yourself.	Exception tuple (à la sys.exc_info)
or, if no exception has occurred, None.
filename	%(filename)s	Filename portion of pathname.
funcName	%(funcName)s	Name of function containing the logging call.
levelname	%(levelname)s	Text logging level for the message ('DEBUG', 'INFO', 'WARNING',
'ERROR', 'CRITICAL').
levelno	%(levelno)s	Numeric logging level for the message (DEBUG, INFO, WARNING, ERROR, CRITICAL).
lineno	%(lineno)d	Source line number where the logging call was issued (if available).
message	%(message)s	The logged message, computed as msg % args. This is set when
Formatter.format() is invoked.
module	%(module)s	Module (name portion of filename).
msecs	%(msecs)d	Millisecond portion of the time when the LogRecord was created.
msg	You shouldn’t need to format this yourself.	The format string passed in the original
logging call. Merged with args to produce message, or an arbitrary object (see Using arbitrary
objects as messages).
name	%(name)s	Name of the logger used to log the call.
pathname	%(pathname)s	Full pathname of the source file where the logging call was issued
(if available).
process	%(process)d	Process ID (if available).
processName	%(processName)s	Process name (if available).
relativeCreated	%(relativeCreated)d	Time in milliseconds when the LogRecord was created,
relative to the time the logging module was loaded.
stack_info	You shouldn’t need to format this yourself.	Stack frame information
(where available) from the bottom of the stack in the current thread, up to and including
the stack frame of the logging call which resulted in the creation of this record.
thread	%(thread)d	Thread ID (if available).
threadName	%(threadName)s	Thread name (if available).




The problem here is that you're not initializing the root logger; you're initializing the logger
for your main module.

Try this for main.py:

import logging
from logging.handlers import RotatingFileHandler
import submodule

logger = logging.getLogger()  # Gets the root logger
logger.setLevel(logging.DEBUG)

fh = RotatingFileHandler('master.log', maxBytes=2000000, backupCount=10)
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
fh.setFormatter(formatter)
logger.addHandler(fh)

logger.debug('DEBUG LEVEL - MAIN MODULE')
logger.info('INFO LEVEL - MAIN MODULE')

submodule.loggerCall()
Then try this for submodule.py:

def loggerCall():
    logger = logging.getLogger(__name__)
    logger.debug('SUBMODULE: DEBUG LOGGING MODE : ')
    logger.info('Submodule: INFO LOG')
    return
Since you said you wanted to send log messages from all your submodules to the same place, you
should initialize the root logger and then simply use the message logging methods
(along with setlevel() calls, as appropriate). Because there's no explicit handler for
your submodule, logging.getLogger(__name__) will traverse the tree to the root,
where it will find the handler you established in main.py.



"""
import logging
import os
import random
import socket
import sys
from logging.handlers import TimedRotatingFileHandler

# import arrow
import datetime

################### Logs #################################################################
APP_ID = __file__ + "," + str(os.getpid()) + "," + str(socket.gethostname())
APP_ID2 = str(os.getpid()) + "_" + str(socket.gethostname())

LOG_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)), "logfile.log")
FORMATTER_1 = logging.Formatter("%(asctime)s,  %(name)s, %(levelname)s, %(message)s")
FORMATTER_2 = logging.Formatter("%(asctime)s.%(msecs)03dZ %(levelname)s %(message)s")
FORMATTER_3 = logging.Formatter("%(asctime)s  %(levelname)s %(message)s")
FORMATTER_4 = logging.Formatter("%(asctime)s, %(process)d, %(filename)s,    %(message)s")


FORMATTER_5 = logging.Formatter(
    "%(asctime)s, %(process)d, %(pathname)s%(filename)s, %(funcName)s, %(lineno)s,  %(message)s"
)


# LOG_FILE = "my_app.log"


#########################################################################################
def create_appid(filename):
    # appid  = filename + ',' + str(os.getpid()) + ',' + str( socket.gethostname() )
    appid = filename + "," + str(os.getpid())
    return appid


def create_logfilename(filename):
    return filename.split("/")[-1].split(".")[0] + ".log"


def create_uniqueid():
    return datetime.datetime.now().strftime(  "_%Y%m%d%H%M%S_"  )   + str(random.randint(1000, 9999))
    # return arrow.utcnow().to("Japan").format("_YYYYMMDDHHmmss_") + str(random.randint(1000, 9999))


########################################################################################
################### Logger #############################################################
def logger_setup(
    logger_name=None,
    log_file=None,
    formatter=FORMATTER_1,
    isrotate=False,
    isconsole_output=True,
    logging_level=logging.DEBUG,
):
    """
    my_logger = util_log.logger_setup("my module name", log_file="")
    APP_ID    = util_log.create_appid(__file__ )
    def log(*argv):
      my_logger.info(",".join([str(x) for x in argv]))
  
   """

    if logger_name is None:
        logger = logging.getLogger()  # Gets the root logger
    else:
        logger = logging.getLogger(logger_name)

    logger.setLevel(logging_level)  # better to have too much log than not enough

    if isconsole_output:
        logger.addHandler(logger_handler_console(formatter))

    if log_file is not None:
        logger.addHandler(
            logger_handler_file(formatter=formatter, log_file_used=log_file, isrotate=isrotate)
        )

    # with this pattern, it's rarely necessary to propagate the error up to parent
    logger.propagate = False
    return logger


def logger_handler_console(formatter=None):
    formatter = FORMATTER_1 if formatter is None else formatter
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(formatter)
    return console_handler


def logger_handler_file(isrotate=False, rotate_time="midnight", formatter=None, log_file_used=None):
    formatter = FORMATTER_1 if formatter is None else formatter
    log_file_used = LOG_FILE if log_file_used is None else log_file_used
    if isrotate:
        print("Rotate log", rotate_time)
        fh = TimedRotatingFileHandler(log_file_used, when=rotate_time)
        fh.setFormatter(formatter)
        return fh
    else:
        fh = logging.FileHandler(log_file_used)
        fh.setFormatter(formatter)
        return fh


def logger_setup2(name=__name__, level=None):
    _ = level

    # logger defines
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    formatter = logging.Formatter()
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    return logger


##########################################################################################
################### Print ################################################################
def printlog(
    s="",
    s1="",
    s2="",
    s3="",
    s4="",
    s5="",
    s6="",
    s7="",
    s8="",
    s9="",
    s10="",
    app_id="",
    logfile=None,
    iswritelog=True,
):
    try:
        if app_id != "":
            prefix = app_id + "," + datetime.datetime.now().strftime(  "_%Y%m%d%H%M%S_"  ) 
        else:
            prefix = APP_ID + "," + datetime.datetime.now().strftime(  "_%Y%m%d%H%M%S_"  ) 
        s = ",".join(
            [
                prefix,
                str(s),
                str(s1),
                str(s2),
                str(s3),
                str(s4),
                str(s5),
                str(s6),
                str(s7),
                str(s8),
                str(s9),
                str(s10),
            ]
        )

        print(s)
        if writelog:
            writelog(s, logfile)
    except Exception as e:
        print(e)
        if iswritelog:
            writelog(str(e), logfile)


def writelog(m="", f=None):
    f = LOG_FILE if f is None else f
    with open(f, "a") as _log:
        _log.write(m + "\n")



def load_arguments(config_file=None, arg_list=None):
    """
      Load CLI input, load config.toml , overwrite config.toml by CLI Input
      [{}, {}]
    """
    import toml
    import argparse
    
    if config_file is not None :
      cur_path = os.path.dirname(os.path.realpath(__file__))
      config_file = os.path.join(cur_path, "config.toml")


    p = argparse.ArgumentParser()
    p.add_argument("--config_file", default=config_file, help="Params File")
    p.add_argument("--config_mode", default="test", help=" test/ prod /uat")
    for x in arg_list : 
        p.add_argument(x["--"], default=x.get("default"), help=x.get("help") )

    arg = p.parse_args()
    
    
    # Load file params as dict namespace
    class to_name(object):
        def __init__(self, adict):
            self.__dict__.update(adict)

    print(arg.config_file)
    pars = toml.load(arg.config_file)
    pars = pars[arg.config_mode]       # test / prod
    print(arg.config_file, pars)

    ### Overwrite params by CLI input and merge with toml file
    for key, x in vars(arg).items():
        if x is not None:  # only values NOT set by CLI
            pars[key] = x

    # print(pars)
    pars = to_name(pars)  #  like object/namespace pars.instance
    return pars



"""A setup module for autoscaleaws."""


# setuptools for distribution
from setuptools import find_packages, setup
import os


with open('requirements.txt') as f:
    required = f.read().splitlines()

LONG_DESCRIPTION = """
 Autoscale AWS allows to scale up virtual machines by invoking spot requests as per
 start and stop rules.
"""

setup(
    name='autoscaleaws',
    version='0.0.2',
    description='Auto scale AWS by spot instances.',
    long_description=LONG_DESCRIPTION,
    license = "BSD",
    # The project's main homepage.
    url='https://github.com/ajeybk/autoscaleaws',
    # Author(s) details
    author='arita37',
    classifiers=[
        "Development Status :: 3 - Alpha",
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
        "License :: OSI Approved :: BSD License",
    ],
    packages=find_packages(where="src", exclude=[]),
    package_dir={'': 'src'},
    install_requires=required,
    python_requires='>=3.0',
    entry_points={
        'console_scripts': [
            'autoscale = autoscale_aws.batch_daemon_autoscale_cli:autoscale_main'
        ]
    }
)

